import json
from typing import Any, Dict, Iterable, List, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import unix_time
from moto.moto_api._internal import mock_random
from moto.utilities.tagging_service import TaggingService


class FilterResource(BaseModel):
    def __init__(
        self,
        region: str,
        account_id: str,
        name: str,
        reason: Optional[str],
        action: str,
        description: Optional[str],
        filter_criteria: Dict[str, Any],
        backend: "Inspector2Backend",
    ):
        filter_id = mock_random.get_random_hex(10)
        self.owner_id = account_id
        self.arn = f"arn:aws:inspector2:{region}:{account_id}:owner/{self.owner_id}/filter/{filter_id}"
        self.name = name
        self.reason = reason
        self.action = action
        self.description = description
        self.filter_criteria = filter_criteria
        self.created_at = unix_time()
        self.backend = backend

    def to_json(self) -> Dict[str, Any]:
        return {
            "action": self.action,
            "arn": self.arn,
            "createdAt": self.created_at,
            "criteria": self.filter_criteria,
            "description": self.description,
            "name": self.name,
            "ownerId": self.owner_id,
            "reason": self.reason,
            "tags": self.backend.list_tags_for_resource(self.arn),
        }


class AccountStatus(BaseModel):
    def __init__(self, account_id: str):
        self.account_id = account_id
        self.ec2 = "DISABLED"
        self.ecr = "DISABLED"
        self._lambda = "DISABLED"
        self.lambda_code = "DISABLED"

    def toggle(self, resource_types: List[str], enable: bool) -> None:
        if "EC2" in resource_types:
            self.ec2 = "ENABLED" if enable else "DISABLED"
        if "ECR" in resource_types:
            self.ecr = "ENABLED" if enable else "DISABLED"
        if "LAMBDA" in resource_types:
            self._lambda = "ENABLED" if enable else "DISABLED"
        if "LAMBDA_CODE" in resource_types or "LAMBDACODE" in resource_types:
            self.lambda_code = "ENABLED" if enable else "DISABLED"

    def to_json(self) -> Dict[str, Any]:
        return {
            "accountId": self.account_id,
            "resourceStatus": {
                "ec2": self.ec2,
                "ecr": self.ecr,
                "lambda": self._lambda,
                "lambdaCode": self.lambda_code,
            },
            "status": self._status(),
        }

    def _status(self) -> str:
        return (
            "ENABLED"
            if "ENABLED" in [self.ec2, self.ecr, self._lambda, self.lambda_code]
            else "DISABLED"
        )

    def to_batch_json(self) -> Dict[str, Any]:
        return {
            "accountId": self.account_id,
            "resourceState": {
                "ec2": {"status": self.ec2},
                "ecr": {"status": self.ecr},
                "lambda": {"status": self._lambda},
                "lambdaCode": {"status": self.lambda_code},
            },
            "state": {"status": self._status()},
        }


class Member(BaseModel):
    def __init__(self, account_id: str, admin_account_id: str):
        self.account_id = account_id
        self.admin_account_id = admin_account_id
        self.status = "ENABLED"
        self.updated_at = unix_time()

    def to_json(self) -> Dict[str, Any]:
        return {
            "accountId": self.account_id,
            "delegatedAdminAccountId": self.admin_account_id,
            "relationshipStatus": self.status,
            "updatedAt": self.updated_at,
        }


class Inspector2Backend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.filters: Dict[str, FilterResource] = dict()
        self.admin_accounts: Dict[str, str] = dict()
        self.account_status: Dict[str, AccountStatus] = dict()
        self.members: Dict[str, Member] = dict()
        self.org_config = {
            "ec2": False,
            "ecr": False,
            "lambda": False,
            "lambdaCode": False,
        }
        self.tagger = TaggingService()
        self.findings_queue: List[Any] = []
        self.findings: Dict[str, Any] = {}

    def create_filter(
        self,
        action: str,
        description: str,
        filter_criteria: Dict[str, Any],
        name: str,
        reason: str,
        tags: Dict[str, str],
    ) -> str:
        _filter = FilterResource(
            region=self.region_name,
            account_id=self.account_id,
            action=action,
            description=description,
            filter_criteria=filter_criteria,
            name=name,
            reason=reason,
            backend=self,
        )
        self.filters[_filter.arn] = _filter
        self.tag_resource(_filter.arn, tags)
        return _filter.arn

    def delete_filter(self, arn: str) -> None:
        self.filters.pop(arn, None)

    def list_filters(self, action: str, arns: List[str]) -> Iterable[FilterResource]:
        """
        Pagination is not yet implemented
        """
        return [
            f
            for f in self.filters.values()
            if (arns and f.arn in arns)
            or (action and f.action == action)
            or (not arns and not action)
        ]

    def list_findings(
        self,
        filter_criteria: List[Dict[str, Any]],
        max_results: str,
        next_token: str,
        sort_criteria: str,
    ) -> List[Dict[str, Any]]:
        """
        This call will always return 0 findings by default.

        You can use a dedicated API to override this, by configuring a queue of expected results.

        A request to `list_findings` will take the first result from that queue, and assign it to the provided arguments. Subsequent calls using the same arguments will return the same result. Other requests using a different SQL-query will take the next result from the queue, or return an empty result if the queue is empty.

        Configure this queue by making an HTTP request to `/moto-api/static/inspector2/findings-results`. An example invocation looks like this:

        .. sourcecode:: python

            findings = {
                "results": [
                    [{
                        "awsAccountId": "111122223333",
                        "codeVulnerabilityDetails": {"cwes": ["a"], "detectorId": ".."},
                    }],
                    # .. other findings as required
                ],
                "account_id": "123456789012",  # This is the default - can be omitted
                "region": "us-east-1",  # This is the default - can be omitted
            }
            resp = requests.post(
                "http://motoapi.amazonaws.com/moto-api/static/inspector2/findings-results",
                json=findings,
            )

            inspector2 = boto3.client("inspector2", region_name="us-east-1")
            findings = inspector2.list_findings()["findings"]

        """
        key = f"{json.dumps(filter_criteria)}--{max_results}--{next_token}--{sort_criteria}"
        if key not in self.findings and self.findings_queue:
            self.findings[key] = self.findings_queue.pop(0)
        if key in self.findings:
            return self.findings[key]
        else:
            return []

    def list_delegated_admin_accounts(self) -> Dict[str, str]:
        return self.admin_accounts

    def enable_delegated_admin_account(self, account_id: str) -> None:
        self.admin_accounts[account_id] = "ENABLED"

    def disable_delegated_admin_account(self, account_id: str) -> None:
        self.admin_accounts[account_id] = "DISABLED"

    def describe_organization_configuration(self) -> Dict[str, Any]:
        return {"autoEnable": self.org_config, "maxAccountLimitReached": False}

    def update_organization_configuration(
        self, auto_enable: Dict[str, bool]
    ) -> Dict[str, Any]:
        self.org_config.update(auto_enable)
        return {"autoEnable": self.org_config}

    def disable(
        self, account_ids: List[str], resource_types: List[str]
    ) -> List[Dict[str, Any]]:
        for acct in account_ids:
            if acct not in self.account_status:
                self.account_status[acct] = AccountStatus(acct)
            self.account_status[acct].toggle(resource_types, enable=False)

        return [
            status.to_json()
            for a_id, status in self.account_status.items()
            if a_id in account_ids
        ]

    def enable(
        self, account_ids: List[str], resource_types: List[str]
    ) -> List[Dict[str, Any]]:
        for acct in account_ids:
            if acct not in self.account_status:
                self.account_status[acct] = AccountStatus(acct)
            self.account_status[acct].toggle(resource_types, enable=True)

        return [
            status.to_json()
            for a_id, status in self.account_status.items()
            if a_id in account_ids
        ]

    def batch_get_account_status(self, account_ids: List[str]) -> List[Dict[str, Any]]:
        return [
            status.to_batch_json()
            for a_id, status in self.account_status.items()
            if a_id in account_ids
        ]

    def list_members(self) -> Iterable[Member]:
        return self.members.values()

    def associate_member(self, account_id: str) -> None:
        self.members[account_id] = Member(
            account_id=account_id, admin_account_id=self.account_id
        )

    def disassociate_member(self, account_id: str) -> None:
        self.members[account_id].status = "DISABLED"

    def get_member(self, account_id: str) -> Member:
        return self.members[account_id]

    def tag_resource(self, resource_arn: str, tags: Dict[str, str]) -> None:
        self.tagger.tag_resource(
            resource_arn, TaggingService.convert_dict_to_tags_input(tags)
        )

    def list_tags_for_resource(self, resource_arn: str) -> Dict[str, str]:
        return self.tagger.get_tag_dict_for_resource(resource_arn)

    def untag_resource(self, arn: str, tag_keys: List[str]) -> None:
        self.tagger.untag_resource_using_names(arn, tag_keys)


inspector2_backends = BackendDict(Inspector2Backend, "inspector2")



from typing import Any, Optional

from moto.core.exceptions import JsonRESTError


class LogsClientError(JsonRESTError):
    code = 400


class ResourceNotFoundException(LogsClientError):
    def __init__(self, msg: Optional[str] = None):
        self.code = 400
        super().__init__(
            "ResourceNotFoundException",
            msg or "The specified log group does not exist.",
        )


class InvalidParameterException(LogsClientError):
    def __init__(
        self,
        msg: Optional[str] = None,
        constraint: Optional[str] = None,
        parameter: Optional[str] = None,
        value: Any = None,
    ):
        self.code = 400
        if constraint:
            msg = f"1 validation error detected: Value '{value}' at '{parameter}' failed to satisfy constraint: {constraint}"
        super().__init__(
            "InvalidParameterException", msg or "A parameter is specified incorrectly."
        )


class ResourceAlreadyExistsException(LogsClientError):
    def __init__(self) -> None:
        self.code = 400
        super().__init__(
            "ResourceAlreadyExistsException", "The specified log group already exists"
        )


class LimitExceededException(LogsClientError):
    def __init__(self) -> None:
        self.code = 400
        super().__init__("LimitExceededException", "Resource limit exceeded.")



"""servicediscovery base URL and path."""

from .responses import ServiceDiscoveryResponse

url_bases = [
    r"https?://servicediscovery\.(.+)\.amazonaws\.com",
]


url_paths = {
    "{0}/$": ServiceDiscoveryResponse.dispatch,
}



from __future__ import annotations

import abc
from typing import Final, Optional

from moto.stepfunctions.parser.asl.component.common.comment import Comment
from moto.stepfunctions.parser.asl.component.common.flow.start_at import StartAt
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.component.states import States


class IterationComponent(EvalComponent, abc.ABC):
    _start_at: Final[StartAt]
    _states: Final[States]
    _comment: Final[Optional[Comment]]

    def __init__(
        self,
        start_at: StartAt,
        states: States,
        comment: Optional[Comment],
    ):
        self._start_at = start_at
        self._states = states
        self._comment = comment



from .responses import InstanceMetadataResponse

url_bases = ["http://169.254.169.254"]

url_paths = {
    "{0}/(?P<path>.+)": InstanceMetadataResponse.method_dispatch(
        InstanceMetadataResponse.metadata_response
    )
}



from typing import List, Optional, Tuple

from moto.utilities.tokenizer import GenericTokenizer


class ParsedQuery:
    def __init__(self) -> None:
        self.limit: Optional[int] = None
        self.fields: List[str] = []
        self.sort: List[Tuple[str, str]] = []

    def sort_reversed(self) -> bool:
        # Descending is the default
        if self.sort:
            # sort_reversed is True if we want to sort in ascending order
            return self.sort[-1][-1] == "asc"
        return False


def parse_query(query: str) -> ParsedQuery:
    tokenizer = GenericTokenizer(query)
    state = "COMMAND"
    characters = ""
    parsed_query = ParsedQuery()

    for char in tokenizer:
        if char.isspace():
            if state == "SORT":
                parsed_query.sort.append((characters, "desc"))
                characters = ""
                state = "SORT_ORDER"
            if state == "COMMAND":
                if characters.lower() in ["fields", "limit", "sort"]:
                    state = characters.upper()
                else:
                    # Unknown/Unsupported command
                    pass
                characters = ""
            tokenizer.skip_white_space()
            continue

        if char == "|":
            if state == "FIELDS":
                parsed_query.fields.append(characters)
                characters = ""
            if state == "LIMIT":
                parsed_query.limit = int(characters)
                characters = ""
            if state == "SORT_ORDER":
                if characters != "":
                    parsed_query.sort[-1] = (parsed_query.sort[-1][0], characters)
                    characters = ""
            state = "COMMAND"
            tokenizer.skip_white_space()
            continue

        if char == ",":
            if state == "FIELDS":
                parsed_query.fields.append(characters)
                characters = ""
                continue

        characters += char

    if state == "FIELDS":
        parsed_query.fields.append(characters)
    if state == "LIMIT":
        parsed_query.limit = int(characters)
    if state == "SORT":
        parsed_query.sort.append((characters, "desc"))
    if state == "SORT_ORDER":
        parsed_query.sort[-1] = (parsed_query.sort[-1][0], characters)

    return parsed_query



"""Exceptions raised by the emrcontainers service."""

from moto.core.exceptions import JsonRESTError


class ResourceNotFoundException(JsonRESTError):
    code = 400

    def __init__(self, resource: str):
        super().__init__("ResourceNotFoundException", resource)



from moto.ec2.utils import add_tag_specification

from ._base_response import EC2BaseResponse


class NatGateways(EC2BaseResponse):
    def create_nat_gateway(self) -> str:
        subnet_id = self._get_param("SubnetId")
        allocation_id = self._get_param("AllocationId")
        connectivity_type = self._get_param("ConnectivityType")
        tags = add_tag_specification(self._get_multi_param("TagSpecification"))

        nat_gateway = self.ec2_backend.create_nat_gateway(
            subnet_id=subnet_id,
            allocation_id=allocation_id,
            tags=tags,
            connectivity_type=connectivity_type,
        )
        template = self.response_template(CREATE_NAT_GATEWAY)
        return template.render(nat_gateway=nat_gateway)

    def delete_nat_gateway(self) -> str:
        nat_gateway_id = self._get_param("NatGatewayId")
        nat_gateway = self.ec2_backend.delete_nat_gateway(nat_gateway_id)
        template = self.response_template(DELETE_NAT_GATEWAY_RESPONSE)
        return template.render(nat_gateway=nat_gateway)

    def describe_nat_gateways(self) -> str:
        filters = self._filters_from_querystring()
        nat_gateway_ids = self._get_multi_param("NatGatewayId")
        nat_gateways = self.ec2_backend.describe_nat_gateways(filters, nat_gateway_ids)
        template = self.response_template(DESCRIBE_NAT_GATEWAYS_RESPONSE)
        return template.render(nat_gateways=nat_gateways)


DESCRIBE_NAT_GATEWAYS_RESPONSE = """<DescribeNatGatewaysResponse xmlns="http://ec2.amazonaws.com/doc/2015-10-01/">
    <requestId>bfed02c6-dae9-47c0-86a2-example</requestId>
    <natGatewaySet>
    {% for nat_gateway in nat_gateways %}
         <item>
            <subnetId>{{ nat_gateway.subnet_id }}</subnetId>
            <natGatewayAddressSet>
            {% for address_set in nat_gateway.address_set %}
                <item>
                    {% if address_set.allocationId %}
                    <allocationId>{{ address_set.allocationId }}</allocationId>
                    {% endif %}
                    {% if address_set.privateIp %}
                    <privateIp>{{ address_set.privateIp }}</privateIp>
                    {% endif %}
                    {% if address_set.publicIp %}
                    <publicIp>{{ address_set.publicIp }}</publicIp>
                    {% endif %}
                    {% if address_set.networkInterfaceId %}
                    <networkInterfaceId>{{ address_set.networkInterfaceId }}</networkInterfaceId>
                    {% endif %}
                    {% if address_set.associationId %}
                    <associationId>{{ address_set.associationId }}</associationId>
                    {% endif %}
                </item>
            {% endfor %}
            </natGatewayAddressSet>
            <createTime>{{ nat_gateway.create_time }}</createTime>
            <vpcId>{{ nat_gateway.vpc_id }}</vpcId>
            <natGatewayId>{{ nat_gateway.id }}</natGatewayId>
            <connectivityType>{{ nat_gateway.connectivity_type }}</connectivityType>
            <state>{{ nat_gateway.state }}</state>
            <tagSet>
                {% for tag in nat_gateway.get_tags() %}
                <item>
                    <key>{{ tag.key }}</key>
                    <value>{{ tag.value }}</value>
                </item>
                {% endfor %}
            </tagSet>
        </item>
    {% endfor %}
    </natGatewaySet>
</DescribeNatGatewaysResponse>
"""

CREATE_NAT_GATEWAY = """<CreateNatGatewayResponse xmlns="http://ec2.amazonaws.com/doc/2015-10-01/">
    <requestId>1b74dc5c-bcda-403f-867d-example</requestId>
    <natGateway>
        <subnetId>{{ nat_gateway.subnet_id }}</subnetId>
        <natGatewayAddressSet>
            {% for address_set in nat_gateway.address_set %}
                <item>
                    {% if address_set.allocationId %}
                    <allocationId>{{ address_set.allocationId }}</allocationId>
                    {% endif %}
                    {% if address_set.privateIp %}
                    <privateIp>{{ address_set.privateIp }}</privateIp>
                    {% endif %}
                    {% if address_set.publicIp %}
                    <publicIp>{{ address_set.publicIp }}</publicIp>
                    {% endif %}
                    {% if address_set.networkInterfaceId %}
                    <networkInterfaceId>{{ address_set.networkInterfaceId }}</networkInterfaceId>
                    {% endif %}
                    {% if address_set.associationId %}
                    <associationId>{{ address_set.associationId }}</associationId>
                    {% endif %}
                </item>
            {% endfor %}
        </natGatewayAddressSet>
        <createTime>{{ nat_gateway.create_time }}</createTime>
        <vpcId>{{ nat_gateway.vpc_id }}</vpcId>
        <natGatewayId>{{ nat_gateway.id }}</natGatewayId>
        <connectivityType>{{ nat_gateway.connectivity_type }}</connectivityType>
        <state>{{ nat_gateway.state }}</state>
        <tagSet>
        {% for tag in nat_gateway.get_tags() %}
            <item>
                <key>{{ tag.key }}</key>
                <value>{{ tag.value }}</value>
            </item>
        {% endfor %}
        </tagSet>
    </natGateway>
</CreateNatGatewayResponse>
"""


DELETE_NAT_GATEWAY_RESPONSE = """<DeleteNatGatewayResponse xmlns="http://ec2.amazonaws.com/doc/2015-10-01/">
    <requestId>741fc8ab-6ebe-452b-b92b-example</requestId>
    <natGatewayId>{{ nat_gateway.id }}</natGatewayId>
</DeleteNatGatewayResponse>"""



"""ivs base URL and path."""

from .responses import IVSResponse

url_bases = [
    r"https?://ivs\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/CreateChannel": IVSResponse.dispatch,
    "{0}/ListChannels": IVSResponse.dispatch,
    "{0}/GetChannel": IVSResponse.dispatch,
    "{0}/BatchGetChannel": IVSResponse.dispatch,
    "{0}/UpdateChannel": IVSResponse.dispatch,
    "{0}/DeleteChannel": IVSResponse.dispatch,
}



import abc

from moto.stepfunctions.parser.asl.component.component import Component


class FunctionName(Component, abc.ABC):
    name: str

    def __init__(self, name: str):
        self.name = name



import json
from typing import Any, Dict, List

from moto.core.responses import BaseResponse

from .exceptions import AWSValidationException
from .models import (
    ApplicationAutoscalingBackend,
    FakeApplicationAutoscalingPolicy,
    FakeScalableTarget,
    FakeScheduledAction,
    ScalableDimensionValueSet,
    ServiceNamespaceValueSet,
    applicationautoscaling_backends,
)


class ApplicationAutoScalingResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="application-autoscaling")

    @property
    def applicationautoscaling_backend(self) -> ApplicationAutoscalingBackend:
        return applicationautoscaling_backends[self.current_account][self.region]

    def describe_scalable_targets(self) -> str:
        self._validate_params()
        service_namespace = self._get_param("ServiceNamespace")
        resource_ids = self._get_param("ResourceIds")
        scalable_dimension = self._get_param("ScalableDimension")
        max_results = self._get_int_param("MaxResults") or 50
        marker = self._get_param("NextToken")
        all_scalable_targets = (
            self.applicationautoscaling_backend.describe_scalable_targets(
                service_namespace, resource_ids, scalable_dimension
            )
        )
        start = int(marker) + 1 if marker else 0
        next_token = None
        scalable_targets_resp = all_scalable_targets[start : start + max_results]
        if len(all_scalable_targets) > start + max_results:
            next_token = str(len(scalable_targets_resp) - 1)
        targets = [_build_target(t) for t in scalable_targets_resp]
        return json.dumps({"ScalableTargets": targets, "NextToken": next_token})

    def register_scalable_target(self) -> str:
        """Registers or updates a scalable target."""
        self._validate_params()
        self.applicationautoscaling_backend.register_scalable_target(
            self._get_param("ServiceNamespace"),
            self._get_param("ResourceId"),
            self._get_param("ScalableDimension"),
            min_capacity=self._get_int_param("MinCapacity"),
            max_capacity=self._get_int_param("MaxCapacity"),
            role_arn=self._get_param("RoleARN"),
            suspended_state=self._get_param("SuspendedState"),
        )
        return json.dumps({})

    def deregister_scalable_target(self) -> str:
        """Deregisters a scalable target."""
        self._validate_params()
        self.applicationautoscaling_backend.deregister_scalable_target(
            self._get_param("ServiceNamespace"),
            self._get_param("ResourceId"),
            self._get_param("ScalableDimension"),
        )
        return json.dumps({})

    def put_scaling_policy(self) -> str:
        policy = self.applicationautoscaling_backend.put_scaling_policy(
            policy_name=self._get_param("PolicyName"),
            service_namespace=self._get_param("ServiceNamespace"),
            resource_id=self._get_param("ResourceId"),
            scalable_dimension=self._get_param("ScalableDimension"),
            policy_type=self._get_param("PolicyType"),
            policy_body=self._get_param(
                "StepScalingPolicyConfiguration",
                self._get_param("TargetTrackingScalingPolicyConfiguration"),
            ),
        )
        return json.dumps(
            {"PolicyARN": policy.policy_arn, "Alarms": _build_alarms(policy)}
        )

    def describe_scaling_policies(self) -> str:
        (
            next_token,
            policy_page,
        ) = self.applicationautoscaling_backend.describe_scaling_policies(
            service_namespace=self._get_param("ServiceNamespace"),
            resource_id=self._get_param("ResourceId"),
            scalable_dimension=self._get_param("ScalableDimension"),
            max_results=self._get_int_param("MaxResults"),
            next_token=self._get_param("NextToken"),
        )
        response_obj = {
            "ScalingPolicies": [_build_policy(p) for p in policy_page],
            "NextToken": next_token,
        }
        return json.dumps(response_obj)

    def delete_scaling_policy(self) -> str:
        self.applicationautoscaling_backend.delete_scaling_policy(
            policy_name=self._get_param("PolicyName"),
            service_namespace=self._get_param("ServiceNamespace"),
            resource_id=self._get_param("ResourceId"),
            scalable_dimension=self._get_param("ScalableDimension"),
        )
        return json.dumps({})

    def _validate_params(self) -> None:
        """Validate parameters.
        TODO Integrate this validation with the validation in models.py
        """
        namespace = self._get_param("ServiceNamespace")
        dimension = self._get_param("ScalableDimension")
        messages = []
        dimensions = [d.value for d in ScalableDimensionValueSet]
        message = None
        if dimension is not None and dimension not in dimensions:
            messages.append(
                f"Value '{dimension}' at 'scalableDimension' failed to satisfy constraint: Member must satisfy enum value set: {dimensions}"
            )
        namespaces = [n.value for n in ServiceNamespaceValueSet]
        if namespace is not None and namespace not in namespaces:
            messages.append(
                f"Value '{namespace}' at 'serviceNamespace' failed to satisfy constraint: Member must satisfy enum value set: {namespaces}"
            )
        if len(messages) == 1:
            message = f"1 validation error detected: {messages[0]}"
        elif len(messages) > 1:
            message = (
                f'{len(messages)} validation errors detected: {"; ".join(messages)}'
            )
        if message:
            raise AWSValidationException(message)

    def delete_scheduled_action(self) -> str:
        params = json.loads(self.body)
        service_namespace = params.get("ServiceNamespace")
        scheduled_action_name = params.get("ScheduledActionName")
        resource_id = params.get("ResourceId")
        scalable_dimension = params.get("ScalableDimension")
        self.applicationautoscaling_backend.delete_scheduled_action(
            service_namespace=service_namespace,
            scheduled_action_name=scheduled_action_name,
            resource_id=resource_id,
            scalable_dimension=scalable_dimension,
        )
        return json.dumps(dict())

    def put_scheduled_action(self) -> str:
        params = json.loads(self.body)
        service_namespace = params.get("ServiceNamespace")
        schedule = params.get("Schedule")
        timezone = params.get("Timezone")
        scheduled_action_name = params.get("ScheduledActionName")
        resource_id = params.get("ResourceId")
        scalable_dimension = params.get("ScalableDimension")
        start_time = params.get("StartTime")
        end_time = params.get("EndTime")
        scalable_target_action = params.get("ScalableTargetAction")
        self.applicationautoscaling_backend.put_scheduled_action(
            service_namespace=service_namespace,
            schedule=schedule,
            timezone=timezone,
            scheduled_action_name=scheduled_action_name,
            resource_id=resource_id,
            scalable_dimension=scalable_dimension,
            start_time=start_time,
            end_time=end_time,
            scalable_target_action=scalable_target_action,
        )
        return json.dumps(dict())

    def describe_scheduled_actions(self) -> str:
        params = json.loads(self.body)
        scheduled_action_names = params.get("ScheduledActionNames")
        service_namespace = params.get("ServiceNamespace")
        resource_id = params.get("ResourceId")
        scalable_dimension = params.get("ScalableDimension")
        scheduled_actions = (
            self.applicationautoscaling_backend.describe_scheduled_actions(
                scheduled_action_names=scheduled_action_names,
                service_namespace=service_namespace,
                resource_id=resource_id,
                scalable_dimension=scalable_dimension,
            )
        )
        response_obj = {
            "ScheduledActions": [_build_scheduled_action(a) for a in scheduled_actions]
        }
        return json.dumps(response_obj)


def _build_target(t: FakeScalableTarget) -> Dict[str, Any]:
    return {
        "CreationTime": t.creation_time,
        "ServiceNamespace": t.service_namespace,
        "ResourceId": t.resource_id,
        "RoleARN": t.role_arn,
        "ScalableDimension": t.scalable_dimension,
        "MaxCapacity": t.max_capacity,
        "MinCapacity": t.min_capacity,
        "SuspendedState": t.suspended_state,
    }


def _build_alarms(policy: FakeApplicationAutoscalingPolicy) -> List[Dict[str, str]]:
    return [{"AlarmARN": a.alarm_arn, "AlarmName": a.name} for a in policy.alarms]


def _build_policy(p: FakeApplicationAutoscalingPolicy) -> Dict[str, Any]:
    response = {
        "PolicyARN": p.policy_arn,
        "PolicyName": p.policy_name,
        "ServiceNamespace": p.service_namespace,
        "ResourceId": p.resource_id,
        "ScalableDimension": p.scalable_dimension,
        "PolicyType": p.policy_type,
        "CreationTime": p.creation_time,
        "Alarms": _build_alarms(p),
    }
    if p.policy_type == "StepScaling":
        response["StepScalingPolicyConfiguration"] = p.step_scaling_policy_configuration
    elif p.policy_type == "TargetTrackingScaling":
        response["TargetTrackingScalingPolicyConfiguration"] = (
            p.target_tracking_scaling_policy_configuration
        )
    return response


def _build_scheduled_action(a: FakeScheduledAction) -> Dict[str, Any]:
    response = {
        "ScheduledActionName": a.scheduled_action_name,
        "ScheduledActionARN": a.arn,
        "ServiceNamespace": a.service_namespace,
        "Schedule": a.schedule,
        "Timezone": a.timezone,
        "ResourceId": a.resource_id,
        "ScalableDimension": a.scalable_dimension,
        "StartTime": a.start_time,
        "EndTime": a.end_time,
        "CreationTime": a.creation_time,
        "ScalableTargetAction": a.scalable_target_action,
    }
    return response



from typing import Final, List

from moto.stepfunctions.parser.asl.component.common.error_name.error_name import (
    ErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class ErrorEqualsDecl(EvalComponent):
    """
    ErrorEquals value MUST be a non-empty array of Strings, which match Error Names.
    Each Retrier MUST contain a field named "ErrorEquals" whose value MUST be a non-empty array of Strings,
    which match Error Names.
    """

    _STATE_ALL_ERROR: Final[StatesErrorName] = StatesErrorName(
        typ=StatesErrorNameType.StatesALL
    )
    _STATE_TASK_ERROR: Final[StatesErrorName] = StatesErrorName(
        typ=StatesErrorNameType.StatesTaskFailed
    )

    def __init__(self, error_names: List[ErrorName]):
        # The reserved name "States.ALL" in a Retrier’s "ErrorEquals" field is a wildcard
        # and matches any Error Name. Such a value MUST appear alone in the "ErrorEquals"
        # array and MUST appear in the last Retrier in the "Retry" array.
        if ErrorEqualsDecl._STATE_ALL_ERROR in error_names and len(error_names) > 1:
            raise ValueError(
                f"States.ALL must appear alone in the ErrorEquals array, got '{error_names}'."
            )

        # TODO: how to handle duplicate ErrorName?
        self.error_names: List[ErrorName] = error_names

    def _eval_body(self, env: Environment) -> None:
        """
        When a state reports an error, the interpreter scans through the Retriers and,
        when the Error Name appears in the value of a Retrier’s "ErrorEquals" field, implements the retry policy
        described in that Retrier.
        This pops the error from the stack, and appends the bool of this check.
        """

        # Try to reduce error response to ErrorName or pass exception upstream.
        error_name: ErrorName = env.stack.pop()

        if ErrorEqualsDecl._STATE_ALL_ERROR in self.error_names:
            res = True
        elif (
            ErrorEqualsDecl._STATE_TASK_ERROR in self.error_names
            and not isinstance(error_name, StatesErrorName)
        ):  # TODO: consider binding a 'context' variable to error_names to more formally detect their evaluation type.
            res = True
        else:
            res = error_name in self.error_names

        env.stack.append(res)



from .responses import MediaLiveResponse

url_bases = [
    r"https?://medialive\.(.+)\.amazonaws.com",
]


response = MediaLiveResponse()


url_paths = {
    "{0}/prod/channels$": response.dispatch,
    "{0}/prod/channels/(?P<channelid>[^/.]+)$": response.dispatch,
    "{0}/prod/channels/(?P<channelid>[^/.]+)/start$": response.dispatch,
    "{0}/prod/channels/(?P<channelid>[^/.]+)/stop$": response.dispatch,
    "{0}/prod/inputs$": response.dispatch,
    "{0}/prod/inputs/(?P<inputid>[^/.]+)$": response.dispatch,
}



# Generated from /Users/mep/LocalStack/localstack/localstack/services/stepfunctions/asl/antlr/ASLIntrinsicParser.g4 by ANTLR 4.13.1
# encoding: utf-8
import sys
from typing import TextIO

from antlr4 import (
    ATN,
    DFA,
    ATNDeserializer,
    NoViableAltException,
    Parser,
    ParserATNSimulator,
    ParserRuleContext,
    ParseTreeListener,
    ParseTreeVisitor,
    PredictionContextCache,
    RecognitionException,
    RuleContext,
    Token,
    TokenStream,
)


def serializedATN():
    return [
        4,
        1,
        42,
        123,
        2,
        0,
        7,
        0,
        2,
        1,
        7,
        1,
        2,
        2,
        7,
        2,
        2,
        3,
        7,
        3,
        2,
        4,
        7,
        4,
        2,
        5,
        7,
        5,
        2,
        6,
        7,
        6,
        2,
        7,
        7,
        7,
        2,
        8,
        7,
        8,
        2,
        9,
        7,
        9,
        2,
        10,
        7,
        10,
        2,
        11,
        7,
        11,
        2,
        12,
        7,
        12,
        1,
        0,
        1,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        2,
        1,
        2,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        5,
        3,
        40,
        8,
        3,
        10,
        3,
        12,
        3,
        43,
        9,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        3,
        3,
        49,
        8,
        3,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        3,
        4,
        58,
        8,
        4,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        5,
        6,
        68,
        8,
        6,
        10,
        6,
        12,
        6,
        71,
        9,
        6,
        1,
        7,
        1,
        7,
        3,
        7,
        75,
        8,
        7,
        1,
        8,
        1,
        8,
        1,
        9,
        1,
        9,
        1,
        9,
        1,
        10,
        1,
        10,
        1,
        10,
        1,
        10,
        1,
        10,
        1,
        10,
        1,
        10,
        1,
        10,
        1,
        10,
        3,
        10,
        91,
        8,
        10,
        1,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        3,
        11,
        101,
        8,
        11,
        1,
        11,
        1,
        11,
        3,
        11,
        105,
        8,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        4,
        11,
        110,
        8,
        11,
        11,
        11,
        12,
        11,
        111,
        5,
        11,
        114,
        8,
        11,
        10,
        11,
        12,
        11,
        117,
        9,
        11,
        1,
        12,
        1,
        12,
        3,
        12,
        121,
        8,
        12,
        1,
        12,
        0,
        1,
        22,
        13,
        0,
        2,
        4,
        6,
        8,
        10,
        12,
        14,
        16,
        18,
        20,
        22,
        24,
        0,
        4,
        1,
        0,
        20,
        37,
        1,
        0,
        17,
        18,
        2,
        0,
        9,
        10,
        15,
        15,
        1,
        0,
        13,
        14,
        127,
        0,
        26,
        1,
        0,
        0,
        0,
        2,
        28,
        1,
        0,
        0,
        0,
        4,
        33,
        1,
        0,
        0,
        0,
        6,
        48,
        1,
        0,
        0,
        0,
        8,
        57,
        1,
        0,
        0,
        0,
        10,
        59,
        1,
        0,
        0,
        0,
        12,
        62,
        1,
        0,
        0,
        0,
        14,
        74,
        1,
        0,
        0,
        0,
        16,
        76,
        1,
        0,
        0,
        0,
        18,
        78,
        1,
        0,
        0,
        0,
        20,
        90,
        1,
        0,
        0,
        0,
        22,
        104,
        1,
        0,
        0,
        0,
        24,
        120,
        1,
        0,
        0,
        0,
        26,
        27,
        3,
        2,
        1,
        0,
        27,
        1,
        1,
        0,
        0,
        0,
        28,
        29,
        5,
        19,
        0,
        0,
        29,
        30,
        5,
        2,
        0,
        0,
        30,
        31,
        3,
        4,
        2,
        0,
        31,
        32,
        3,
        6,
        3,
        0,
        32,
        3,
        1,
        0,
        0,
        0,
        33,
        34,
        7,
        0,
        0,
        0,
        34,
        5,
        1,
        0,
        0,
        0,
        35,
        36,
        5,
        5,
        0,
        0,
        36,
        41,
        3,
        8,
        4,
        0,
        37,
        38,
        5,
        4,
        0,
        0,
        38,
        40,
        3,
        8,
        4,
        0,
        39,
        37,
        1,
        0,
        0,
        0,
        40,
        43,
        1,
        0,
        0,
        0,
        41,
        39,
        1,
        0,
        0,
        0,
        41,
        42,
        1,
        0,
        0,
        0,
        42,
        44,
        1,
        0,
        0,
        0,
        43,
        41,
        1,
        0,
        0,
        0,
        44,
        45,
        5,
        6,
        0,
        0,
        45,
        49,
        1,
        0,
        0,
        0,
        46,
        47,
        5,
        5,
        0,
        0,
        47,
        49,
        5,
        6,
        0,
        0,
        48,
        35,
        1,
        0,
        0,
        0,
        48,
        46,
        1,
        0,
        0,
        0,
        49,
        7,
        1,
        0,
        0,
        0,
        50,
        58,
        5,
        38,
        0,
        0,
        51,
        58,
        5,
        39,
        0,
        0,
        52,
        58,
        5,
        40,
        0,
        0,
        53,
        58,
        7,
        1,
        0,
        0,
        54,
        58,
        3,
        10,
        5,
        0,
        55,
        58,
        3,
        12,
        6,
        0,
        56,
        58,
        3,
        0,
        0,
        0,
        57,
        50,
        1,
        0,
        0,
        0,
        57,
        51,
        1,
        0,
        0,
        0,
        57,
        52,
        1,
        0,
        0,
        0,
        57,
        53,
        1,
        0,
        0,
        0,
        57,
        54,
        1,
        0,
        0,
        0,
        57,
        55,
        1,
        0,
        0,
        0,
        57,
        56,
        1,
        0,
        0,
        0,
        58,
        9,
        1,
        0,
        0,
        0,
        59,
        60,
        5,
        1,
        0,
        0,
        60,
        61,
        3,
        12,
        6,
        0,
        61,
        11,
        1,
        0,
        0,
        0,
        62,
        63,
        5,
        1,
        0,
        0,
        63,
        64,
        5,
        2,
        0,
        0,
        64,
        69,
        3,
        14,
        7,
        0,
        65,
        66,
        5,
        2,
        0,
        0,
        66,
        68,
        3,
        14,
        7,
        0,
        67,
        65,
        1,
        0,
        0,
        0,
        68,
        71,
        1,
        0,
        0,
        0,
        69,
        67,
        1,
        0,
        0,
        0,
        69,
        70,
        1,
        0,
        0,
        0,
        70,
        13,
        1,
        0,
        0,
        0,
        71,
        69,
        1,
        0,
        0,
        0,
        72,
        75,
        3,
        16,
        8,
        0,
        73,
        75,
        3,
        18,
        9,
        0,
        74,
        72,
        1,
        0,
        0,
        0,
        74,
        73,
        1,
        0,
        0,
        0,
        75,
        15,
        1,
        0,
        0,
        0,
        76,
        77,
        3,
        24,
        12,
        0,
        77,
        17,
        1,
        0,
        0,
        0,
        78,
        79,
        3,
        16,
        8,
        0,
        79,
        80,
        3,
        20,
        10,
        0,
        80,
        19,
        1,
        0,
        0,
        0,
        81,
        82,
        5,
        7,
        0,
        0,
        82,
        91,
        5,
        8,
        0,
        0,
        83,
        84,
        5,
        7,
        0,
        0,
        84,
        85,
        5,
        39,
        0,
        0,
        85,
        91,
        5,
        8,
        0,
        0,
        86,
        87,
        5,
        7,
        0,
        0,
        87,
        88,
        3,
        22,
        11,
        0,
        88,
        89,
        5,
        8,
        0,
        0,
        89,
        91,
        1,
        0,
        0,
        0,
        90,
        81,
        1,
        0,
        0,
        0,
        90,
        83,
        1,
        0,
        0,
        0,
        90,
        86,
        1,
        0,
        0,
        0,
        91,
        21,
        1,
        0,
        0,
        0,
        92,
        93,
        6,
        11,
        -1,
        0,
        93,
        105,
        5,
        3,
        0,
        0,
        94,
        95,
        5,
        11,
        0,
        0,
        95,
        100,
        3,
        16,
        8,
        0,
        96,
        97,
        7,
        2,
        0,
        0,
        97,
        101,
        5,
        39,
        0,
        0,
        98,
        99,
        5,
        16,
        0,
        0,
        99,
        101,
        5,
        38,
        0,
        0,
        100,
        96,
        1,
        0,
        0,
        0,
        100,
        98,
        1,
        0,
        0,
        0,
        101,
        105,
        1,
        0,
        0,
        0,
        102,
        103,
        5,
        12,
        0,
        0,
        103,
        105,
        5,
        39,
        0,
        0,
        104,
        92,
        1,
        0,
        0,
        0,
        104,
        94,
        1,
        0,
        0,
        0,
        104,
        102,
        1,
        0,
        0,
        0,
        105,
        115,
        1,
        0,
        0,
        0,
        106,
        109,
        10,
        1,
        0,
        0,
        107,
        108,
        7,
        3,
        0,
        0,
        108,
        110,
        3,
        22,
        11,
        0,
        109,
        107,
        1,
        0,
        0,
        0,
        110,
        111,
        1,
        0,
        0,
        0,
        111,
        109,
        1,
        0,
        0,
        0,
        111,
        112,
        1,
        0,
        0,
        0,
        112,
        114,
        1,
        0,
        0,
        0,
        113,
        106,
        1,
        0,
        0,
        0,
        114,
        117,
        1,
        0,
        0,
        0,
        115,
        113,
        1,
        0,
        0,
        0,
        115,
        116,
        1,
        0,
        0,
        0,
        116,
        23,
        1,
        0,
        0,
        0,
        117,
        115,
        1,
        0,
        0,
        0,
        118,
        121,
        5,
        41,
        0,
        0,
        119,
        121,
        3,
        4,
        2,
        0,
        120,
        118,
        1,
        0,
        0,
        0,
        120,
        119,
        1,
        0,
        0,
        0,
        121,
        25,
        1,
        0,
        0,
        0,
        11,
        41,
        48,
        57,
        69,
        74,
        90,
        100,
        104,
        111,
        115,
        120,
    ]


class ASLIntrinsicParser(Parser):
    grammarFileName = "ASLIntrinsicParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = [
        "<INVALID>",
        "'$'",
        "'.'",
        "'*'",
        "','",
        "'('",
        "')'",
        "'['",
        "']'",
        "'<'",
        "'>'",
        "'@.'",
        "'@.length-'",
        "'&&'",
        "'||'",
        "'=='",
        "'='",
        "'true'",
        "'false'",
        "'States'",
        "'Format'",
        "'StringToJson'",
        "'JsonToString'",
        "'Array'",
        "'ArrayPartition'",
        "'ArrayContains'",
        "'ArrayRange'",
        "'ArrayGetItem'",
        "'ArrayLength'",
        "'ArrayUnique'",
        "'Base64Encode'",
        "'Base64Decode'",
        "'Hash'",
        "'JsonMerge'",
        "'MathRandom'",
        "'MathAdd'",
        "'StringSplit'",
        "'UUID'",
    ]

    symbolicNames = [
        "<INVALID>",
        "DOLLAR",
        "DOT",
        "STAR",
        "COMMA",
        "LPAREN",
        "RPAREN",
        "LBRACK",
        "RBRACK",
        "LDIAM",
        "RDIAM",
        "ATDOT",
        "ATDOTLENGTHDASH",
        "ANDAND",
        "OROR",
        "EQEQ",
        "EQ",
        "TRUE",
        "FALSE",
        "States",
        "Format",
        "StringToJson",
        "JsonToString",
        "Array",
        "ArrayPartition",
        "ArrayContains",
        "ArrayRange",
        "ArrayGetItem",
        "ArrayLength",
        "ArrayUnique",
        "Base64Encode",
        "Base64Decode",
        "Hash",
        "JsonMerge",
        "MathRandom",
        "MathAdd",
        "StringSplit",
        "UUID",
        "STRING",
        "INT",
        "NUMBER",
        "IDENTIFIER",
        "WS",
    ]

    RULE_func_decl = 0
    RULE_states_func_decl = 1
    RULE_state_fun_name = 2
    RULE_func_arg_list = 3
    RULE_func_arg = 4
    RULE_context_path = 5
    RULE_json_path = 6
    RULE_json_path_part = 7
    RULE_json_path_iden = 8
    RULE_json_path_iden_qual = 9
    RULE_json_path_qual = 10
    RULE_json_path_query = 11
    RULE_identifier = 12

    ruleNames = [
        "func_decl",
        "states_func_decl",
        "state_fun_name",
        "func_arg_list",
        "func_arg",
        "context_path",
        "json_path",
        "json_path_part",
        "json_path_iden",
        "json_path_iden_qual",
        "json_path_qual",
        "json_path_query",
        "identifier",
    ]

    EOF = Token.EOF
    DOLLAR = 1
    DOT = 2
    STAR = 3
    COMMA = 4
    LPAREN = 5
    RPAREN = 6
    LBRACK = 7
    RBRACK = 8
    LDIAM = 9
    RDIAM = 10
    ATDOT = 11
    ATDOTLENGTHDASH = 12
    ANDAND = 13
    OROR = 14
    EQEQ = 15
    EQ = 16
    TRUE = 17
    FALSE = 18
    States = 19
    Format = 20
    StringToJson = 21
    JsonToString = 22
    Array = 23
    ArrayPartition = 24
    ArrayContains = 25
    ArrayRange = 26
    ArrayGetItem = 27
    ArrayLength = 28
    ArrayUnique = 29
    Base64Encode = 30
    Base64Decode = 31
    Hash = 32
    JsonMerge = 33
    MathRandom = 34
    MathAdd = 35
    StringSplit = 36
    UUID = 37
    STRING = 38
    INT = 39
    NUMBER = 40
    IDENTIFIER = 41
    WS = 42

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.1")
        self._interp = ParserATNSimulator(
            self, self.atn, self.decisionsToDFA, self.sharedContextCache
        )
        self._predicates = None

    class Func_declContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def states_func_decl(self):
            return self.getTypedRuleContext(
                ASLIntrinsicParser.States_func_declContext, 0
            )

        def getRuleIndex(self):
            return ASLIntrinsicParser.RULE_func_decl

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunc_decl"):
                listener.enterFunc_decl(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunc_decl"):
                listener.exitFunc_decl(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunc_decl"):
                return visitor.visitFunc_decl(self)
            else:
                return visitor.visitChildren(self)

    def func_decl(self):
        localctx = ASLIntrinsicParser.Func_declContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_func_decl)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 26
            self.states_func_decl()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class States_func_declContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def States(self):
            return self.getToken(ASLIntrinsicParser.States, 0)

        def DOT(self):
            return self.getToken(ASLIntrinsicParser.DOT, 0)

        def state_fun_name(self):
            return self.getTypedRuleContext(ASLIntrinsicParser.State_fun_nameContext, 0)

        def func_arg_list(self):
            return self.getTypedRuleContext(ASLIntrinsicParser.Func_arg_listContext, 0)

        def getRuleIndex(self):
            return ASLIntrinsicParser.RULE_states_func_decl

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStates_func_decl"):
                listener.enterStates_func_decl(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStates_func_decl"):
                listener.exitStates_func_decl(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStates_func_decl"):
                return visitor.visitStates_func_decl(self)
            else:
                return visitor.visitChildren(self)

    def states_func_decl(self):
        localctx = ASLIntrinsicParser.States_func_declContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 2, self.RULE_states_func_decl)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 28
            self.match(ASLIntrinsicParser.States)
            self.state = 29
            self.match(ASLIntrinsicParser.DOT)
            self.state = 30
            self.state_fun_name()
            self.state = 31
            self.func_arg_list()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class State_fun_nameContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Format(self):
            return self.getToken(ASLIntrinsicParser.Format, 0)

        def StringToJson(self):
            return self.getToken(ASLIntrinsicParser.StringToJson, 0)

        def JsonToString(self):
            return self.getToken(ASLIntrinsicParser.JsonToString, 0)

        def Array(self):
            return self.getToken(ASLIntrinsicParser.Array, 0)

        def ArrayPartition(self):
            return self.getToken(ASLIntrinsicParser.ArrayPartition, 0)

        def ArrayContains(self):
            return self.getToken(ASLIntrinsicParser.ArrayContains, 0)

        def ArrayRange(self):
            return self.getToken(ASLIntrinsicParser.ArrayRange, 0)

        def ArrayGetItem(self):
            return self.getToken(ASLIntrinsicParser.ArrayGetItem, 0)

        def ArrayLength(self):
            return self.getToken(ASLIntrinsicParser.ArrayLength, 0)

        def ArrayUnique(self):
            return self.getToken(ASLIntrinsicParser.ArrayUnique, 0)

        def Base64Encode(self):
            return self.getToken(ASLIntrinsicParser.Base64Encode, 0)

        def Base64Decode(self):
            return self.getToken(ASLIntrinsicParser.Base64Decode, 0)

        def Hash(self):
            return self.getToken(ASLIntrinsicParser.Hash, 0)

        def JsonMerge(self):
            return self.getToken(ASLIntrinsicParser.JsonMerge, 0)

        def MathRandom(self):
            return self.getToken(ASLIntrinsicParser.MathRandom, 0)

        def MathAdd(self):
            return self.getToken(ASLIntrinsicParser.MathAdd, 0)

        def StringSplit(self):
            return self.getToken(ASLIntrinsicParser.StringSplit, 0)

        def UUID(self):
            return self.getToken(ASLIntrinsicParser.UUID, 0)

        def getRuleIndex(self):
            return ASLIntrinsicParser.RULE_state_fun_name

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterState_fun_name"):
                listener.enterState_fun_name(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitState_fun_name"):
                listener.exitState_fun_name(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitState_fun_name"):
                return visitor.visitState_fun_name(self)
            else:
                return visitor.visitChildren(self)

    def state_fun_name(self):
        localctx = ASLIntrinsicParser.State_fun_nameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_state_fun_name)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 33
            _la = self._input.LA(1)
            if not (((_la) & ~0x3F) == 0 and ((1 << _la) & 274876858368) != 0):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Func_arg_listContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LPAREN(self):
            return self.getToken(ASLIntrinsicParser.LPAREN, 0)

        def func_arg(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(ASLIntrinsicParser.Func_argContext)
            else:
                return self.getTypedRuleContext(ASLIntrinsicParser.Func_argContext, i)

        def RPAREN(self):
            return self.getToken(ASLIntrinsicParser.RPAREN, 0)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(ASLIntrinsicParser.COMMA)
            else:
                return self.getToken(ASLIntrinsicParser.COMMA, i)

        def getRuleIndex(self):
            return ASLIntrinsicParser.RULE_func_arg_list

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunc_arg_list"):
                listener.enterFunc_arg_list(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunc_arg_list"):
                listener.exitFunc_arg_list(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunc_arg_list"):
                return visitor.visitFunc_arg_list(self)
            else:
                return visitor.visitChildren(self)

    def func_arg_list(self):
        localctx = ASLIntrinsicParser.Func_arg_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_func_arg_list)
        self._la = 0  # Token type
        try:
            self.state = 48
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 1, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 35
                self.match(ASLIntrinsicParser.LPAREN)
                self.state = 36
                self.func_arg()
                self.state = 41
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 4:
                    self.state = 37
                    self.match(ASLIntrinsicParser.COMMA)
                    self.state = 38
                    self.func_arg()
                    self.state = 43
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 44
                self.match(ASLIntrinsicParser.RPAREN)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 46
                self.match(ASLIntrinsicParser.LPAREN)
                self.state = 47
                self.match(ASLIntrinsicParser.RPAREN)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Func_argContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return ASLIntrinsicParser.RULE_func_arg

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class Func_arg_context_pathContext(Func_argContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Func_argContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def context_path(self):
            return self.getTypedRuleContext(ASLIntrinsicParser.Context_pathContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunc_arg_context_path"):
                listener.enterFunc_arg_context_path(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunc_arg_context_path"):
                listener.exitFunc_arg_context_path(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunc_arg_context_path"):
                return visitor.visitFunc_arg_context_path(self)
            else:
                return visitor.visitChildren(self)

    class Func_arg_floatContext(Func_argContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Func_argContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NUMBER(self):
            return self.getToken(ASLIntrinsicParser.NUMBER, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunc_arg_float"):
                listener.enterFunc_arg_float(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunc_arg_float"):
                listener.exitFunc_arg_float(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunc_arg_float"):
                return visitor.visitFunc_arg_float(self)
            else:
                return visitor.visitChildren(self)

    class Func_arg_func_declContext(Func_argContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Func_argContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def func_decl(self):
            return self.getTypedRuleContext(ASLIntrinsicParser.Func_declContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunc_arg_func_decl"):
                listener.enterFunc_arg_func_decl(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunc_arg_func_decl"):
                listener.exitFunc_arg_func_decl(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunc_arg_func_decl"):
                return visitor.visitFunc_arg_func_decl(self)
            else:
                return visitor.visitChildren(self)

    class Func_arg_intContext(Func_argContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Func_argContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def INT(self):
            return self.getToken(ASLIntrinsicParser.INT, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunc_arg_int"):
                listener.enterFunc_arg_int(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunc_arg_int"):
                listener.exitFunc_arg_int(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunc_arg_int"):
                return visitor.visitFunc_arg_int(self)
            else:
                return visitor.visitChildren(self)

    class Func_arg_boolContext(Func_argContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Func_argContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def TRUE(self):
            return self.getToken(ASLIntrinsicParser.TRUE, 0)

        def FALSE(self):
            return self.getToken(ASLIntrinsicParser.FALSE, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunc_arg_bool"):
                listener.enterFunc_arg_bool(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunc_arg_bool"):
                listener.exitFunc_arg_bool(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunc_arg_bool"):
                return visitor.visitFunc_arg_bool(self)
            else:
                return visitor.visitChildren(self)

    class Func_arg_stringContext(Func_argContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Func_argContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def STRING(self):
            return self.getToken(ASLIntrinsicParser.STRING, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunc_arg_string"):
                listener.enterFunc_arg_string(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunc_arg_string"):
                listener.exitFunc_arg_string(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunc_arg_string"):
                return visitor.visitFunc_arg_string(self)
            else:
                return visitor.visitChildren(self)

    class Func_arg_json_pathContext(Func_argContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Func_argContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def json_path(self):
            return self.getTypedRuleContext(ASLIntrinsicParser.Json_pathContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunc_arg_json_path"):
                listener.enterFunc_arg_json_path(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunc_arg_json_path"):
                listener.exitFunc_arg_json_path(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunc_arg_json_path"):
                return visitor.visitFunc_arg_json_path(self)
            else:
                return visitor.visitChildren(self)

    def func_arg(self):
        localctx = ASLIntrinsicParser.Func_argContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_func_arg)
        self._la = 0  # Token type
        try:
            self.state = 57
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 2, self._ctx)
            if la_ == 1:
                localctx = ASLIntrinsicParser.Func_arg_stringContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 50
                self.match(ASLIntrinsicParser.STRING)
                pass

            elif la_ == 2:
                localctx = ASLIntrinsicParser.Func_arg_intContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 51
                self.match(ASLIntrinsicParser.INT)
                pass

            elif la_ == 3:
                localctx = ASLIntrinsicParser.Func_arg_floatContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 52
                self.match(ASLIntrinsicParser.NUMBER)
                pass

            elif la_ == 4:
                localctx = ASLIntrinsicParser.Func_arg_boolContext(self, localctx)
                self.enterOuterAlt(localctx, 4)
                self.state = 53
                _la = self._input.LA(1)
                if not (_la == 17 or _la == 18):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                pass

            elif la_ == 5:
                localctx = ASLIntrinsicParser.Func_arg_context_pathContext(
                    self, localctx
                )
                self.enterOuterAlt(localctx, 5)
                self.state = 54
                self.context_path()
                pass

            elif la_ == 6:
                localctx = ASLIntrinsicParser.Func_arg_json_pathContext(self, localctx)
                self.enterOuterAlt(localctx, 6)
                self.state = 55
                self.json_path()
                pass

            elif la_ == 7:
                localctx = ASLIntrinsicParser.Func_arg_func_declContext(self, localctx)
                self.enterOuterAlt(localctx, 7)
                self.state = 56
                self.func_decl()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Context_pathContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DOLLAR(self):
            return self.getToken(ASLIntrinsicParser.DOLLAR, 0)

        def json_path(self):
            return self.getTypedRuleContext(ASLIntrinsicParser.Json_pathContext, 0)

        def getRuleIndex(self):
            return ASLIntrinsicParser.RULE_context_path

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterContext_path"):
                listener.enterContext_path(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitContext_path"):
                listener.exitContext_path(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitContext_path"):
                return visitor.visitContext_path(self)
            else:
                return visitor.visitChildren(self)

    def context_path(self):
        localctx = ASLIntrinsicParser.Context_pathContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_context_path)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 59
            self.match(ASLIntrinsicParser.DOLLAR)
            self.state = 60
            self.json_path()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Json_pathContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DOLLAR(self):
            return self.getToken(ASLIntrinsicParser.DOLLAR, 0)

        def DOT(self, i: int = None):
            if i is None:
                return self.getTokens(ASLIntrinsicParser.DOT)
            else:
                return self.getToken(ASLIntrinsicParser.DOT, i)

        def json_path_part(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    ASLIntrinsicParser.Json_path_partContext
                )
            else:
                return self.getTypedRuleContext(
                    ASLIntrinsicParser.Json_path_partContext, i
                )

        def getRuleIndex(self):
            return ASLIntrinsicParser.RULE_json_path

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJson_path"):
                listener.enterJson_path(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJson_path"):
                listener.exitJson_path(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitJson_path"):
                return visitor.visitJson_path(self)
            else:
                return visitor.visitChildren(self)

    def json_path(self):
        localctx = ASLIntrinsicParser.Json_pathContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_json_path)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 62
            self.match(ASLIntrinsicParser.DOLLAR)
            self.state = 63
            self.match(ASLIntrinsicParser.DOT)
            self.state = 64
            self.json_path_part()
            self.state = 69
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 2:
                self.state = 65
                self.match(ASLIntrinsicParser.DOT)
                self.state = 66
                self.json_path_part()
                self.state = 71
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Json_path_partContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def json_path_iden(self):
            return self.getTypedRuleContext(ASLIntrinsicParser.Json_path_idenContext, 0)

        def json_path_iden_qual(self):
            return self.getTypedRuleContext(
                ASLIntrinsicParser.Json_path_iden_qualContext, 0
            )

        def getRuleIndex(self):
            return ASLIntrinsicParser.RULE_json_path_part

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJson_path_part"):
                listener.enterJson_path_part(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJson_path_part"):
                listener.exitJson_path_part(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitJson_path_part"):
                return visitor.visitJson_path_part(self)
            else:
                return visitor.visitChildren(self)

    def json_path_part(self):
        localctx = ASLIntrinsicParser.Json_path_partContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_json_path_part)
        try:
            self.state = 74
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 4, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 72
                self.json_path_iden()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 73
                self.json_path_iden_qual()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Json_path_idenContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def identifier(self):
            return self.getTypedRuleContext(ASLIntrinsicParser.IdentifierContext, 0)

        def getRuleIndex(self):
            return ASLIntrinsicParser.RULE_json_path_iden

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJson_path_iden"):
                listener.enterJson_path_iden(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJson_path_iden"):
                listener.exitJson_path_iden(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitJson_path_iden"):
                return visitor.visitJson_path_iden(self)
            else:
                return visitor.visitChildren(self)

    def json_path_iden(self):
        localctx = ASLIntrinsicParser.Json_path_idenContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_json_path_iden)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 76
            self.identifier()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Json_path_iden_qualContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def json_path_iden(self):
            return self.getTypedRuleContext(ASLIntrinsicParser.Json_path_idenContext, 0)

        def json_path_qual(self):
            return self.getTypedRuleContext(ASLIntrinsicParser.Json_path_qualContext, 0)

        def getRuleIndex(self):
            return ASLIntrinsicParser.RULE_json_path_iden_qual

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJson_path_iden_qual"):
                listener.enterJson_path_iden_qual(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJson_path_iden_qual"):
                listener.exitJson_path_iden_qual(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitJson_path_iden_qual"):
                return visitor.visitJson_path_iden_qual(self)
            else:
                return visitor.visitChildren(self)

    def json_path_iden_qual(self):
        localctx = ASLIntrinsicParser.Json_path_iden_qualContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 18, self.RULE_json_path_iden_qual)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 78
            self.json_path_iden()
            self.state = 79
            self.json_path_qual()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Json_path_qualContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return ASLIntrinsicParser.RULE_json_path_qual

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class Json_path_qual_voidContext(Json_path_qualContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Json_path_qualContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LBRACK(self):
            return self.getToken(ASLIntrinsicParser.LBRACK, 0)

        def RBRACK(self):
            return self.getToken(ASLIntrinsicParser.RBRACK, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJson_path_qual_void"):
                listener.enterJson_path_qual_void(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJson_path_qual_void"):
                listener.exitJson_path_qual_void(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitJson_path_qual_void"):
                return visitor.visitJson_path_qual_void(self)
            else:
                return visitor.visitChildren(self)

    class Json_path_qual_queryContext(Json_path_qualContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Json_path_qualContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LBRACK(self):
            return self.getToken(ASLIntrinsicParser.LBRACK, 0)

        def json_path_query(self):
            return self.getTypedRuleContext(
                ASLIntrinsicParser.Json_path_queryContext, 0
            )

        def RBRACK(self):
            return self.getToken(ASLIntrinsicParser.RBRACK, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJson_path_qual_query"):
                listener.enterJson_path_qual_query(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJson_path_qual_query"):
                listener.exitJson_path_qual_query(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitJson_path_qual_query"):
                return visitor.visitJson_path_qual_query(self)
            else:
                return visitor.visitChildren(self)

    class Json_path_qual_idxContext(Json_path_qualContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Json_path_qualContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LBRACK(self):
            return self.getToken(ASLIntrinsicParser.LBRACK, 0)

        def INT(self):
            return self.getToken(ASLIntrinsicParser.INT, 0)

        def RBRACK(self):
            return self.getToken(ASLIntrinsicParser.RBRACK, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJson_path_qual_idx"):
                listener.enterJson_path_qual_idx(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJson_path_qual_idx"):
                listener.exitJson_path_qual_idx(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitJson_path_qual_idx"):
                return visitor.visitJson_path_qual_idx(self)
            else:
                return visitor.visitChildren(self)

    def json_path_qual(self):
        localctx = ASLIntrinsicParser.Json_path_qualContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_json_path_qual)
        try:
            self.state = 90
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 5, self._ctx)
            if la_ == 1:
                localctx = ASLIntrinsicParser.Json_path_qual_voidContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 81
                self.match(ASLIntrinsicParser.LBRACK)
                self.state = 82
                self.match(ASLIntrinsicParser.RBRACK)
                pass

            elif la_ == 2:
                localctx = ASLIntrinsicParser.Json_path_qual_idxContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 83
                self.match(ASLIntrinsicParser.LBRACK)
                self.state = 84
                self.match(ASLIntrinsicParser.INT)
                self.state = 85
                self.match(ASLIntrinsicParser.RBRACK)
                pass

            elif la_ == 3:
                localctx = ASLIntrinsicParser.Json_path_qual_queryContext(
                    self, localctx
                )
                self.enterOuterAlt(localctx, 3)
                self.state = 86
                self.match(ASLIntrinsicParser.LBRACK)
                self.state = 87
                self.json_path_query(0)
                self.state = 88
                self.match(ASLIntrinsicParser.RBRACK)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Json_path_queryContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return ASLIntrinsicParser.RULE_json_path_query

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class Json_path_query_cmpContext(Json_path_queryContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Json_path_queryContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ATDOT(self):
            return self.getToken(ASLIntrinsicParser.ATDOT, 0)

        def json_path_iden(self):
            return self.getTypedRuleContext(ASLIntrinsicParser.Json_path_idenContext, 0)

        def INT(self):
            return self.getToken(ASLIntrinsicParser.INT, 0)

        def EQ(self):
            return self.getToken(ASLIntrinsicParser.EQ, 0)

        def STRING(self):
            return self.getToken(ASLIntrinsicParser.STRING, 0)

        def LDIAM(self):
            return self.getToken(ASLIntrinsicParser.LDIAM, 0)

        def RDIAM(self):
            return self.getToken(ASLIntrinsicParser.RDIAM, 0)

        def EQEQ(self):
            return self.getToken(ASLIntrinsicParser.EQEQ, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJson_path_query_cmp"):
                listener.enterJson_path_query_cmp(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJson_path_query_cmp"):
                listener.exitJson_path_query_cmp(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitJson_path_query_cmp"):
                return visitor.visitJson_path_query_cmp(self)
            else:
                return visitor.visitChildren(self)

    class Json_path_query_lengthContext(Json_path_queryContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Json_path_queryContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ATDOTLENGTHDASH(self):
            return self.getToken(ASLIntrinsicParser.ATDOTLENGTHDASH, 0)

        def INT(self):
            return self.getToken(ASLIntrinsicParser.INT, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJson_path_query_length"):
                listener.enterJson_path_query_length(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJson_path_query_length"):
                listener.exitJson_path_query_length(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitJson_path_query_length"):
                return visitor.visitJson_path_query_length(self)
            else:
                return visitor.visitChildren(self)

    class Json_path_query_binaryContext(Json_path_queryContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Json_path_queryContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def json_path_query(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    ASLIntrinsicParser.Json_path_queryContext
                )
            else:
                return self.getTypedRuleContext(
                    ASLIntrinsicParser.Json_path_queryContext, i
                )

        def ANDAND(self, i: int = None):
            if i is None:
                return self.getTokens(ASLIntrinsicParser.ANDAND)
            else:
                return self.getToken(ASLIntrinsicParser.ANDAND, i)

        def OROR(self, i: int = None):
            if i is None:
                return self.getTokens(ASLIntrinsicParser.OROR)
            else:
                return self.getToken(ASLIntrinsicParser.OROR, i)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJson_path_query_binary"):
                listener.enterJson_path_query_binary(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJson_path_query_binary"):
                listener.exitJson_path_query_binary(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitJson_path_query_binary"):
                return visitor.visitJson_path_query_binary(self)
            else:
                return visitor.visitChildren(self)

    class Json_path_query_starContext(Json_path_queryContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ASLIntrinsicParser.Json_path_queryContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def STAR(self):
            return self.getToken(ASLIntrinsicParser.STAR, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJson_path_query_star"):
                listener.enterJson_path_query_star(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJson_path_query_star"):
                listener.exitJson_path_query_star(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitJson_path_query_star"):
                return visitor.visitJson_path_query_star(self)
            else:
                return visitor.visitChildren(self)

    def json_path_query(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = ASLIntrinsicParser.Json_path_queryContext(
            self, self._ctx, _parentState
        )
        _prevctx = localctx
        _startState = 22
        self.enterRecursionRule(localctx, 22, self.RULE_json_path_query, _p)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 104
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [3]:
                localctx = ASLIntrinsicParser.Json_path_query_starContext(
                    self, localctx
                )
                self._ctx = localctx
                _prevctx = localctx

                self.state = 93
                self.match(ASLIntrinsicParser.STAR)
                pass
            elif token in [11]:
                localctx = ASLIntrinsicParser.Json_path_query_cmpContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 94
                self.match(ASLIntrinsicParser.ATDOT)
                self.state = 95
                self.json_path_iden()
                self.state = 100
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [9, 10, 15]:
                    self.state = 96
                    _la = self._input.LA(1)
                    if not (((_la) & ~0x3F) == 0 and ((1 << _la) & 34304) != 0):
                        self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()
                    self.state = 97
                    self.match(ASLIntrinsicParser.INT)
                    pass
                elif token in [16]:
                    self.state = 98
                    self.match(ASLIntrinsicParser.EQ)
                    self.state = 99
                    self.match(ASLIntrinsicParser.STRING)
                    pass
                else:
                    raise NoViableAltException(self)

                pass
            elif token in [12]:
                localctx = ASLIntrinsicParser.Json_path_query_lengthContext(
                    self, localctx
                )
                self._ctx = localctx
                _prevctx = localctx
                self.state = 102
                self.match(ASLIntrinsicParser.ATDOTLENGTHDASH)
                self.state = 103
                self.match(ASLIntrinsicParser.INT)
                pass
            else:
                raise NoViableAltException(self)

            self._ctx.stop = self._input.LT(-1)
            self.state = 115
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 9, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = ASLIntrinsicParser.Json_path_query_binaryContext(
                        self,
                        ASLIntrinsicParser.Json_path_queryContext(
                            self, _parentctx, _parentState
                        ),
                    )
                    self.pushNewRecursionContext(
                        localctx, _startState, self.RULE_json_path_query
                    )
                    self.state = 106
                    if not self.precpred(self._ctx, 1):
                        from antlr4.error.Errors import FailedPredicateException

                        raise FailedPredicateException(
                            self, "self.precpred(self._ctx, 1)"
                        )
                    self.state = 109
                    self._errHandler.sync(self)
                    _alt = 1
                    while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                        if _alt == 1:
                            self.state = 107
                            _la = self._input.LA(1)
                            if not (_la == 13 or _la == 14):
                                self._errHandler.recoverInline(self)
                            else:
                                self._errHandler.reportMatch(self)
                                self.consume()
                            self.state = 108
                            self.json_path_query(0)

                        else:
                            raise NoViableAltException(self)
                        self.state = 111
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input, 8, self._ctx)

                self.state = 117
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 9, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class IdentifierContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(ASLIntrinsicParser.IDENTIFIER, 0)

        def state_fun_name(self):
            return self.getTypedRuleContext(ASLIntrinsicParser.State_fun_nameContext, 0)

        def getRuleIndex(self):
            return ASLIntrinsicParser.RULE_identifier

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterIdentifier"):
                listener.enterIdentifier(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitIdentifier"):
                listener.exitIdentifier(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitIdentifier"):
                return visitor.visitIdentifier(self)
            else:
                return visitor.visitChildren(self)

    def identifier(self):
        localctx = ASLIntrinsicParser.IdentifierContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_identifier)
        try:
            self.state = 120
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [41]:
                self.enterOuterAlt(localctx, 1)
                self.state = 118
                self.match(ASLIntrinsicParser.IDENTIFIER)
                pass
            elif token in [
                20,
                21,
                22,
                23,
                24,
                25,
                26,
                27,
                28,
                29,
                30,
                31,
                32,
                33,
                34,
                35,
                36,
                37,
            ]:
                self.enterOuterAlt(localctx, 2)
                self.state = 119
                self.state_fun_name()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    def sempred(self, localctx: RuleContext, ruleIndex: int, actionIndex: int):
        if self._predicates is None:
            self._predicates = dict()
        self._predicates[11] = self.json_path_query_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, actionIndex)

    def json_path_query_sempred(self, localctx: Json_path_queryContext, predIndex: int):
        if predIndex == 0:
            return self.precpred(self._ctx, 1)



from typing import TYPE_CHECKING, Any, Dict, List

if TYPE_CHECKING:
    from ..models import LogEvent, LogGroup, LogStream

from .query_parser import ParsedQuery, parse_query


class ParsedEvent:
    def __init__(
        self,
        event: "LogEvent",
        query: ParsedQuery,
        log_stream: "LogStream",
        log_group: "LogGroup",
    ):
        self.event = event
        self.query = query
        self.log_stream = log_stream
        self.log_group = log_group
        self.fields = self._create_fields()

    def _create_fields(self) -> Dict[str, Any]:
        fields: Dict[str, Any] = {"@ptr": self.event.event_id}
        if "@timestamp" in self.query.fields:
            fields["@timestamp"] = self.event.timestamp
        if "@message" in self.query.fields:
            fields["@message"] = self.event.message
        if "@logStream" in self.query.fields:
            fields["@logStream"] = self.log_stream.log_stream_name  # type: ignore[has-type]
        if "@log" in self.query.fields:
            fields["@log"] = self.log_group.name
        return fields

    def __eq__(self, other: "ParsedEvent") -> bool:  # type: ignore[override]
        return self.event.timestamp == other.event.timestamp

    def __lt__(self, other: "ParsedEvent") -> bool:
        return self.event.timestamp < other.event.timestamp

    def __le__(self, other: "ParsedEvent") -> bool:
        return self.event.timestamp <= other.event.timestamp

    def __gt__(self, other: "ParsedEvent") -> bool:
        return self.event.timestamp > other.event.timestamp

    def __ge__(self, other: "ParsedEvent") -> bool:
        return self.event.timestamp >= other.event.timestamp


def execute_query(
    log_groups: List["LogGroup"], query: str, start_time: int, end_time: int
) -> List[Dict[str, str]]:
    parsed = parse_query(query)
    all_events = _create_parsed_events(log_groups, parsed, start_time, end_time)
    sorted_events = sorted(all_events, reverse=parsed.sort_reversed())
    sorted_fields = [event.fields for event in sorted_events]
    if parsed.limit:
        return sorted_fields[0 : parsed.limit]
    return sorted_fields


def _create_parsed_events(
    log_groups: List["LogGroup"], query: ParsedQuery, start_time: int, end_time: int
) -> List["ParsedEvent"]:
    def filter_func(event: "LogEvent") -> bool:
        # Start/End time is in epoch seconds
        # Event timestamp is in epoch milliseconds
        if start_time and event.timestamp < (start_time * 1000):
            return False

        if end_time and event.timestamp > (end_time * 1000):
            return False

        return True

    events: List["ParsedEvent"] = []
    for group in log_groups:
        for stream in group.streams.values():
            events.extend(
                [
                    ParsedEvent(
                        event=event, query=query, log_stream=stream, log_group=group
                    )
                    for event in filter(filter_func, stream.events)
                ]
            )

    return events



from typing import Any, Dict, List, Optional

from moto.core.utils import iso_8601_datetime_with_milliseconds, utcnow

from ..exceptions import (
    InvalidKeyPairDuplicateError,
    InvalidKeyPairFormatError,
    InvalidKeyPairNameError,
)
from ..utils import (
    generic_filter,
    public_key_fingerprint,
    public_key_parse,
    random_ed25519_key_pair,
    random_key_pair_id,
    random_rsa_key_pair,
)
from .core import TaggedEC2Resource


class KeyPair(TaggedEC2Resource):
    def __init__(
        self,
        name: str,
        fingerprint: str,
        material: str,
        tags: Dict[str, str],
        ec2_backend: Any,
    ):
        self.id = random_key_pair_id()
        self.name = name
        self.fingerprint = fingerprint
        self.material = material
        self.create_time = utcnow()
        self.ec2_backend = ec2_backend
        self.add_tags(tags or {})

    @property
    def created_iso_8601(self) -> str:
        return iso_8601_datetime_with_milliseconds(self.create_time)

    def get_filter_value(
        self, filter_name: str, method_name: Optional[str] = None
    ) -> str:
        if filter_name == "key-name":
            return self.name
        elif filter_name == "fingerprint":
            return self.fingerprint
        else:
            return super().get_filter_value(filter_name, "DescribeKeyPairs")


class KeyPairBackend:
    def __init__(self) -> None:
        self.keypairs: Dict[str, KeyPair] = {}

    def create_key_pair(
        self, name: str, key_type: str, tags: Dict[str, str]
    ) -> KeyPair:
        if name in self.keypairs:
            raise InvalidKeyPairDuplicateError(name)
        if key_type == "ed25519":
            keypair = KeyPair(
                name, **random_ed25519_key_pair(), tags=tags, ec2_backend=self
            )
        else:
            keypair = KeyPair(
                name, **random_rsa_key_pair(), tags=tags, ec2_backend=self
            )

        self.keypairs[name] = keypair
        return keypair

    def delete_key_pair(self, name: str) -> None:
        self.keypairs.pop(name, None)

    def describe_key_pairs(
        self, key_names: List[str], filters: Any = None
    ) -> List[KeyPair]:
        if any(key_names):
            results = [
                keypair
                for keypair in self.keypairs.values()
                if keypair.name in key_names
            ]
            if len(key_names) > len(results):
                unknown_keys = set(key_names) - set(results)  # type: ignore
                raise InvalidKeyPairNameError(unknown_keys)
        else:
            results = list(self.keypairs.values())

        if filters:
            return generic_filter(filters, results)
        else:
            return results

    def import_key_pair(
        self, key_name: str, public_key_material: str, tags: Dict[str, str]
    ) -> KeyPair:
        if key_name in self.keypairs:
            raise InvalidKeyPairDuplicateError(key_name)

        try:
            public_key = public_key_parse(public_key_material)
        except ValueError:
            raise InvalidKeyPairFormatError()

        fingerprint = public_key_fingerprint(public_key)
        keypair = KeyPair(
            key_name,
            material=public_key_material,
            fingerprint=fingerprint,
            tags=tags,
            ec2_backend=self,
        )
        self.keypairs[key_name] = keypair
        return keypair



from moto.core.exceptions import JsonRESTError


class ConflictException(JsonRESTError):
    def __init__(self, message: str):
        super().__init__("ConflictException", message)


class BadRequestException(JsonRESTError):
    def __init__(self, message: str):
        super().__init__("BadRequestException", message)



from typing import Any, Dict, Final, Optional, Set

from botocore.exceptions import ClientError

from moto.stepfunctions.parser.api import HistoryEventType, TaskFailedEventDetails
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ResourceRuntimePart,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_callback import (
    StateTaskServiceCallback,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.boto_client import boto_client_for
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str


class StateTaskServiceSqs(StateTaskServiceCallback):
    _ERROR_NAME_CLIENT: Final[str] = "SQS.SdkClientException"
    _ERROR_NAME_AWS: Final[str] = "SQS.AmazonSQSException"

    _SUPPORTED_API_PARAM_BINDINGS: Final[Dict[str, Set[str]]] = {
        "sendmessage": {
            "DelaySeconds",
            "MessageAttribute",
            "MessageBody",
            "MessageDeduplicationId",
            "MessageGroupId",
            "QueueUrl",
        }
    }

    def _get_supported_parameters(self) -> Optional[Set[str]]:
        return self._SUPPORTED_API_PARAM_BINDINGS.get(self.resource.api_action.lower())

    def _from_error(self, env: Environment, ex: Exception) -> FailureEvent:
        if isinstance(ex, ClientError):
            return FailureEvent(
                error_name=CustomErrorName(self._ERROR_NAME_CLIENT),
                event_type=HistoryEventType.TaskFailed,
                event_details=EventDetails(
                    taskFailedEventDetails=TaskFailedEventDetails(
                        error=self._ERROR_NAME_CLIENT,
                        cause=ex.response["Error"][
                            "Message"
                        ],  # TODO: update to report expected cause.
                        resource=self._get_sfn_resource(),
                        resourceType=self._get_sfn_resource_type(),
                    )
                ),
            )
        return super()._from_error(env=env, ex=ex)

    def _normalise_response(
        self,
        response: Any,
        boto_service_name: Optional[str] = None,
        service_action_name: Optional[str] = None,
    ) -> None:
        super()._normalise_response(
            response=response,
            boto_service_name=boto_service_name,
            service_action_name=service_action_name,
        )
        # Normalise output value key to SFN standard for Md5OfMessageBody.
        if response and "Md5OfMessageBody" in response:
            md5_message_body = response.pop("Md5OfMessageBody")
            response["MD5OfMessageBody"] = md5_message_body

    def _eval_service_task(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
    ):
        # TODO: Stepfunctions automatically dumps to json MessageBody's definitions.
        #  Are these other similar scenarios?
        if "MessageBody" in normalised_parameters:
            message_body = normalised_parameters["MessageBody"]
            if message_body is not None and not isinstance(message_body, str):
                normalised_parameters["MessageBody"] = to_json_str(message_body)

        service_name = self._get_boto_service_name()
        api_action = self._get_boto_service_action()
        sqs_client = boto_client_for(
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            service=service_name,
        )
        response = getattr(sqs_client, api_action)(**normalised_parameters)
        response.pop("ResponseMetadata", None)
        env.stack.append(response)



from typing import Final

from moto.stepfunctions.parser.asl.component.component import Component


class MaxAttemptsDecl(Component):
    """
    "MaxAttempts": value MUST be a non-negative integer, representing the maximum number
    of retry attempts (default: 3)
    """

    DEFAULT_ATTEMPTS: Final[int] = 3

    def __init__(self, attempts: int = DEFAULT_ATTEMPTS):
        if attempts < 0:
            raise ValueError(
                f"MaxAttempts value MUST be a non-negative integer, got '{attempts}'."
            )
        self.attempts: Final[int] = attempts



from typing import Any

from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import BaseResponse

from ... import recorder


class RecorderResponse(BaseResponse):
    def reset_recording(
        self,
        req: Any,
        url: str,
        headers: Any,  # pylint: disable=unused-argument
    ) -> TYPE_RESPONSE:
        recorder.reset_recording()
        return 200, {}, ""

    def start_recording(
        self,
        req: Any,
        url: str,
        headers: Any,  # pylint: disable=unused-argument
    ) -> TYPE_RESPONSE:
        recorder.start_recording()
        return 200, {}, "Recording is set to True"

    def stop_recording(
        self,
        req: Any,
        url: str,
        headers: Any,  # pylint: disable=unused-argument
    ) -> TYPE_RESPONSE:
        recorder.stop_recording()
        return 200, {}, "Recording is set to False"

    def upload_recording(
        self,
        req: Any,
        url: str,
        headers: Any,  # pylint: disable=unused-argument
    ) -> TYPE_RESPONSE:
        data = req.data
        recorder.upload_recording(data)
        return 200, {}, ""

    def download_recording(
        self,
        req: Any,
        url: str,
        headers: Any,  # pylint: disable=unused-argument
    ) -> TYPE_RESPONSE:
        data = recorder.download_recording()
        return 200, {}, data

    # NOTE: Replaying assumes, for simplicity, that it is the only action
    # running against moto at the time. No recording happens while replaying.
    def replay_recording(
        self,
        req: Any,
        url: str,
        headers: Any,  # pylint: disable=unused-argument
    ) -> TYPE_RESPONSE:
        recorder.replay_recording(target_host=url)
        return 200, {}, ""



from moto import settings

from .responses import S3Response

# Catch s3.amazonaws.com, but not s3-control.amazonaws.com
url_bases = [
    r"https?://s3(?!-control)(.*)\.amazonaws.com",
    r"https?://(?P<bucket_name>[a-zA-Z0-9\-_.]*)\.?s3(?!-control)(.*)\.amazonaws.com",
]

url_bases.extend(settings.get_s3_custom_endpoints())

url_paths = {
    # subdomain bucket
    "{0}/$": S3Response.method_dispatch(S3Response.bucket_response),
    # subdomain key of path-based bucket
    "{0}/(?P<key_or_bucket_name>[^/]+)$": S3Response.method_dispatch(
        S3Response.ambiguous_response
    ),
    "{0}/(?P<key_or_bucket_name>[^/]+)/$": S3Response.method_dispatch(
        S3Response.ambiguous_response
    ),
    # path-based bucket + key
    "{0}/(?P<bucket_name_path>[^/]+)/(?P<key_name>.+)": S3Response.method_dispatch(
        S3Response.key_response
    ),
    # subdomain bucket + key with empty first part of path
    "{0}/(?P<key_name>/.*)$": S3Response.method_dispatch(S3Response.key_response),
}



from typing import Tuple

from moto.core.base_backend import BackendDict, BaseBackend
from moto.kinesisvideo.models import KinesisVideoBackend, kinesisvideo_backends
from moto.sts.utils import random_session_token


class KinesisVideoArchivedMediaBackend(BaseBackend):
    @property
    def backend(self) -> KinesisVideoBackend:
        return kinesisvideo_backends[self.account_id][self.region_name]

    def _get_streaming_url(
        self, stream_name: str, stream_arn: str, api_name: str
    ) -> str:
        stream = self.backend._get_stream(stream_name, stream_arn)
        data_endpoint = stream.get_data_endpoint(api_name)
        session_token = random_session_token()
        api_to_relative_path = {
            "GET_HLS_STREAMING_SESSION_URL": "/hls/v1/getHLSMasterPlaylist.m3u8",
            "GET_DASH_STREAMING_SESSION_URL": "/dash/v1/getDASHManifest.mpd",
        }
        relative_path = api_to_relative_path[api_name]
        return f"{data_endpoint}{relative_path}?SessionToken={session_token}"

    def get_hls_streaming_session_url(self, stream_name: str, stream_arn: str) -> str:
        # Ignore option parameters as the format of hls_url doesn't depend on them
        api_name = "GET_HLS_STREAMING_SESSION_URL"
        return self._get_streaming_url(stream_name, stream_arn, api_name)

    def get_dash_streaming_session_url(self, stream_name: str, stream_arn: str) -> str:
        # Ignore option parameters as the format of hls_url doesn't depend on them
        api_name = "GET_DASH_STREAMING_SESSION_URL"
        return self._get_streaming_url(stream_name, stream_arn, api_name)

    def get_clip(self, stream_name: str, stream_arn: str) -> Tuple[str, bytes]:
        self.backend._get_stream(stream_name, stream_arn)
        content_type = "video/mp4"  # Fixed content_type as it depends on input stream
        payload = b"sample-mp4-video"
        return content_type, payload


kinesisvideoarchivedmedia_backends = BackendDict(
    KinesisVideoArchivedMediaBackend, "kinesis-video-archived-media"
)



import copy
import logging
from typing import Final, Optional

from moto.stepfunctions.parser.asl.component.common.parameters import Parameters
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iteration_worker import (
    IterationWorker,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.job import (
    JobPool,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment

LOG = logging.getLogger(__name__)


class IteratorWorker(IterationWorker):
    _parameters: Final[Optional[Parameters]]

    def __init__(
        self,
        work_name: str,
        job_pool: JobPool,
        env: Environment,
        parameters: Optional[Parameters],
    ):
        super().__init__(work_name=work_name, job_pool=job_pool, env=env)
        self._parameters = parameters

    def _eval_input(self, env_frame: Environment) -> None:
        if self._parameters:
            map_state_input = self._env.stack[-1]
            env_frame.inp = copy.deepcopy(map_state_input)
            env_frame.stack.append(env_frame.inp)
            self._parameters.eval(env_frame)
            env_frame.inp = env_frame.stack.pop()
            env_frame.stack.append(env_frame.inp)



from .responses import ResourceGroupsResponse

url_bases = [r"https?://resource-groups(-fips)?\.(.+)\.amazonaws.com"]

url_paths = {
    "{0}/delete-group$": ResourceGroupsResponse.dispatch,
    "{0}/get-group$": ResourceGroupsResponse.dispatch,
    "{0}/get-group-configuration$": ResourceGroupsResponse.dispatch,
    "{0}/put-group-configuration$": ResourceGroupsResponse.dispatch,
    "{0}/get-group-query$": ResourceGroupsResponse.dispatch,
    "{0}/groups$": ResourceGroupsResponse.dispatch,
    "{0}/groups/(?P<resource_group_name>[^/]+)$": ResourceGroupsResponse.dispatch,
    "{0}/groups/(?P<resource_group_name>[^/]+)/query$": ResourceGroupsResponse.dispatch,
    "{0}/groups-list$": ResourceGroupsResponse.dispatch,
    "{0}/resources/(?P<resource_arn>[^/]+)/tags$": ResourceGroupsResponse.dispatch,
    "{0}/update-group$": ResourceGroupsResponse.dispatch,
    "{0}/update-group-query$": ResourceGroupsResponse.dispatch,
}



from moto.stepfunctions.parser.asl.component.intrinsic.functionname.function_name import (
    FunctionName,
)
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.state_function_name_types import (
    StatesFunctionNameType,
)


class StatesFunctionName(FunctionName):
    def __init__(self, function_type: StatesFunctionNameType):
        super().__init__(name=function_type.name())
        self.function_type: StatesFunctionNameType = function_type



from typing import Final, List

from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payload_value import (
    PayloadValue,
)
from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payloadbinding.payload_binding import (
    PayloadBinding,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class PayloadTmpl(PayloadValue):
    def __init__(self, payload_bindings: List[PayloadBinding]):
        self.payload_bindings: Final[List[PayloadBinding]] = payload_bindings

    def _eval_body(self, env: Environment) -> None:
        env.stack.append(dict())
        for payload_binding in self.payload_bindings:
            payload_binding.eval(env)



from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument import (
    FunctionArgument,
)


class FunctionArgumentBool(FunctionArgument):
    _value: bool

    def __init__(self, boolean: bool):
        super().__init__(value=boolean)



from typing import Optional

from moto.stepfunctions.parser.api import HistoryEventType, TaskFailedEventDetails
from moto.stepfunctions.parser.asl.component.common.cause_decl import CauseDecl
from moto.stepfunctions.parser.asl.component.common.error_decl import ErrorDecl
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.state.state import CommonStateField
from moto.stepfunctions.parser.asl.component.state.state_props import StateProps
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails


class StateFail(CommonStateField):
    def __init__(self):
        super().__init__(
            state_entered_event_type=HistoryEventType.FailStateEntered,
            state_exited_event_type=None,
        )
        self.cause: Optional[CauseDecl] = None
        self.error: Optional[ErrorDecl] = None

    def from_state_props(self, state_props: StateProps) -> None:
        super(StateFail, self).from_state_props(state_props)
        self.cause = state_props.get(CauseDecl)
        self.error = state_props.get(ErrorDecl)

    def _eval_state(self, env: Environment) -> None:
        task_failed_event_details = TaskFailedEventDetails()
        if self.error:
            task_failed_event_details["error"] = self.error.error
        if self.cause:
            task_failed_event_details["cause"] = self.cause.cause

        error_name = CustomErrorName(self.error.error) if self.error else None
        failure_event = FailureEvent(
            error_name=error_name,
            event_type=HistoryEventType.TaskFailed,
            event_details=EventDetails(
                taskFailedEventDetails=task_failed_event_details
            ),
        )
        raise FailureEventException(failure_event=failure_event)



# import json
import string
from typing import Any, List, Optional, Tuple

from moto.moto_api._internal import mock_random as random


def get_partition(region: str) -> str:
    valid_matches = [
        # (region prefix, aws partition)
        ("cn-", "aws-cn"),
        ("us-gov-", "aws-us-gov"),
        ("us-gov-iso-", "aws-iso"),
        ("us-gov-iso-b-", "aws-iso-b"),
    ]

    for prefix, partition in valid_matches:
        if region.startswith(prefix):
            return partition
    return "aws"


def random_id(size: int = 13) -> str:
    chars = list(range(10)) + list(string.ascii_lowercase)
    return "".join(str(random.choice(chars)) for x in range(size))


def random_cluster_id() -> str:
    return random_id(size=25)


def random_job_id() -> str:
    return random_id(size=19)


def paginated_list(
    full_list: List[Any], sort_key: str, max_results: int, next_token: Optional[str]
) -> Tuple[List[Any], Optional[str]]:
    """
    Returns a tuple containing a slice of the full list starting at next_token and ending with at most the max_results
    number of elements, and the new next_token which can be passed back in for the next segment of the full list.
    """
    if next_token is None or not next_token:
        next_token = 0  # type: ignore
    next_token = int(next_token)  # type: ignore

    sorted_list = sorted(full_list, key=lambda d: d[sort_key])

    values = sorted_list[next_token : next_token + max_results]  # type: ignore
    if len(values) == max_results:
        new_next_token = str(next_token + max_results)  # type: ignore
    else:
        new_next_token = None
    return values, new_next_token



# Generated from /Users/mep/LocalStack/localstack/localstack/services/stepfunctions/asl/antlr/ASLIntrinsicParser.g4 by ANTLR 4.13.1
from antlr4 import ParseTreeVisitor

if "." in __name__:
    from .ASLIntrinsicParser import ASLIntrinsicParser
else:
    from ASLIntrinsicParser import ASLIntrinsicParser

# This class defines a complete generic visitor for a parse tree produced by ASLIntrinsicParser.


class ASLIntrinsicParserVisitor(ParseTreeVisitor):
    # Visit a parse tree produced by ASLIntrinsicParser#func_decl.
    def visitFunc_decl(self, ctx: ASLIntrinsicParser.Func_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#states_func_decl.
    def visitStates_func_decl(self, ctx: ASLIntrinsicParser.States_func_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#state_fun_name.
    def visitState_fun_name(self, ctx: ASLIntrinsicParser.State_fun_nameContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#func_arg_list.
    def visitFunc_arg_list(self, ctx: ASLIntrinsicParser.Func_arg_listContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#func_arg_string.
    def visitFunc_arg_string(self, ctx: ASLIntrinsicParser.Func_arg_stringContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#func_arg_int.
    def visitFunc_arg_int(self, ctx: ASLIntrinsicParser.Func_arg_intContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#func_arg_float.
    def visitFunc_arg_float(self, ctx: ASLIntrinsicParser.Func_arg_floatContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#func_arg_bool.
    def visitFunc_arg_bool(self, ctx: ASLIntrinsicParser.Func_arg_boolContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#func_arg_context_path.
    def visitFunc_arg_context_path(
        self, ctx: ASLIntrinsicParser.Func_arg_context_pathContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#func_arg_json_path.
    def visitFunc_arg_json_path(
        self, ctx: ASLIntrinsicParser.Func_arg_json_pathContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#func_arg_func_decl.
    def visitFunc_arg_func_decl(
        self, ctx: ASLIntrinsicParser.Func_arg_func_declContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#context_path.
    def visitContext_path(self, ctx: ASLIntrinsicParser.Context_pathContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#json_path.
    def visitJson_path(self, ctx: ASLIntrinsicParser.Json_pathContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#json_path_part.
    def visitJson_path_part(self, ctx: ASLIntrinsicParser.Json_path_partContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#json_path_iden.
    def visitJson_path_iden(self, ctx: ASLIntrinsicParser.Json_path_idenContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#json_path_iden_qual.
    def visitJson_path_iden_qual(
        self, ctx: ASLIntrinsicParser.Json_path_iden_qualContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#json_path_qual_void.
    def visitJson_path_qual_void(
        self, ctx: ASLIntrinsicParser.Json_path_qual_voidContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#json_path_qual_idx.
    def visitJson_path_qual_idx(
        self, ctx: ASLIntrinsicParser.Json_path_qual_idxContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#json_path_qual_query.
    def visitJson_path_qual_query(
        self, ctx: ASLIntrinsicParser.Json_path_qual_queryContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#json_path_query_cmp.
    def visitJson_path_query_cmp(
        self, ctx: ASLIntrinsicParser.Json_path_query_cmpContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#json_path_query_length.
    def visitJson_path_query_length(
        self, ctx: ASLIntrinsicParser.Json_path_query_lengthContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#json_path_query_binary.
    def visitJson_path_query_binary(
        self, ctx: ASLIntrinsicParser.Json_path_query_binaryContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#json_path_query_star.
    def visitJson_path_query_star(
        self, ctx: ASLIntrinsicParser.Json_path_query_starContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLIntrinsicParser#identifier.
    def visitIdentifier(self, ctx: ASLIntrinsicParser.IdentifierContext):
        return self.visitChildren(ctx)


del ASLIntrinsicParser



from typing import Final

from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument import (
    FunctionArgument,
)
from moto.stepfunctions.parser.asl.component.intrinsic.function.function import Function
from moto.stepfunctions.parser.asl.eval.environment import Environment


class FunctionArgumentFunction(FunctionArgument):
    def __init__(self, function: Function):
        super().__init__()
        self.function: Final[Function] = function

    def _eval_body(self, env: Environment) -> None:
        self.function.eval(env=env)
        self._value = env.stack.pop()
        super()._eval_body(env=env)



PAGINATION_MODEL = {
    "list_workspaces": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,  # This should be the sum of the directory limits
        "unique_attribute": "arn",
    },
    "list_rule_groups_namespaces": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,  # This should be the sum of the directory limits
        "unique_attribute": "name",
    },
}



from .responses import GuardDutyResponse

url_bases = [
    "https?://guardduty\\.(.+)\\.amazonaws\\.com",
]


url_paths = {
    "{0}/detector$": GuardDutyResponse.method_dispatch(GuardDutyResponse.detectors),
    "{0}/detector/(?P<detector_id>[^/]+)$": GuardDutyResponse.method_dispatch(
        GuardDutyResponse.detector
    ),
    "{0}/detector/(?P<detector_id>[^/]+)/filter$": GuardDutyResponse.method_dispatch(
        GuardDutyResponse.filters
    ),
    "{0}/detector/(?P<detector_id>[^/]+)/filter/(?P<filter_name>[^/]+)$": GuardDutyResponse.method_dispatch(
        GuardDutyResponse.filter
    ),
    "{0}/admin/enable$": GuardDutyResponse.method_dispatch(
        GuardDutyResponse.enable_organization_admin_account
    ),
    "{0}/admin$": GuardDutyResponse.method_dispatch(
        GuardDutyResponse.list_organization_admin_accounts
    ),
}



"""servicequotas base URL and path."""

from .responses import ServiceQuotasResponse

url_bases = [
    r"https?://servicequotas\.(.+)\.amazonaws\.com",
]


url_paths = {
    "{0}/$": ServiceQuotasResponse.dispatch,
}



import boto3
from botocore.client import BaseClient
from botocore.config import Config

from moto.core.models import botocore_stubber
from moto.settings import moto_server_port
from moto.stepfunctions.parser.asl.component.common.timeouts.timeout import (
    TimeoutSeconds,
)


def boto_client_for(region: str, account: str, service: str) -> BaseClient:
    intercepting_boto_calls = botocore_stubber.enabled
    kwargs = {}
    if not intercepting_boto_calls:
        kwargs["endpoint_url"] = f"http://localhost:{moto_server_port()}"
    return boto3.client(
        aws_access_key_id=account,
        region_name=region,
        service_name=service,
        aws_secret_access_key="sk",
        config=Config(
            parameter_validation=False,
            retries={"max_attempts": 0, "total_max_attempts": 1},
            connect_timeout=TimeoutSeconds.DEFAULT_TIMEOUT_SECONDS,
            read_timeout=TimeoutSeconds.DEFAULT_TIMEOUT_SECONDS,
        ),
        **kwargs,
    )



import re
import time
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Tuple

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import iso_8601_datetime_without_milliseconds, utcnow
from moto.utilities.tagging_service import TaggingService

from .exceptions import (
    InsufficientSnsTopicPolicyException,
    S3BucketDoesNotExistException,
    TrailNameInvalidChars,
    TrailNameNotEndingCorrectly,
    TrailNameNotStartingCorrectly,
    TrailNameTooLong,
    TrailNameTooShort,
    TrailNotFoundException,
)


def datetime2int(date: datetime) -> int:
    return int(time.mktime(date.timetuple()))


class TrailStatus:
    def __init__(self) -> None:
        self.is_logging = False
        self.latest_delivery_time: Optional[int] = None
        self.latest_delivery_attempt: Optional[str] = ""
        self.started: Optional[datetime] = None
        self.stopped: Optional[datetime] = None

    def start_logging(self) -> None:
        self.is_logging = True
        self.started = utcnow()
        self.latest_delivery_time = datetime2int(utcnow())
        self.latest_delivery_attempt = iso_8601_datetime_without_milliseconds(utcnow())

    def stop_logging(self) -> None:
        self.is_logging = False
        self.stopped = utcnow()

    def description(self) -> Dict[str, Any]:
        if self.is_logging:
            self.latest_delivery_time = datetime2int(utcnow())
            self.latest_delivery_attempt = iso_8601_datetime_without_milliseconds(
                utcnow()
            )
        desc: Dict[str, Any] = {
            "IsLogging": self.is_logging,
            "LatestDeliveryAttemptTime": self.latest_delivery_attempt,
            "LatestNotificationAttemptTime": "",
            "LatestNotificationAttemptSucceeded": "",
            "LatestDeliveryAttemptSucceeded": "",
            "TimeLoggingStarted": "",
            "TimeLoggingStopped": "",
        }
        if self.started:
            desc["StartLoggingTime"] = datetime2int(self.started)
            desc["TimeLoggingStarted"] = iso_8601_datetime_without_milliseconds(
                self.started
            )
            desc["LatestDeliveryTime"] = self.latest_delivery_time
        if self.stopped:
            desc["StopLoggingTime"] = datetime2int(self.stopped)
            desc["TimeLoggingStopped"] = iso_8601_datetime_without_milliseconds(
                self.stopped
            )
        return desc


class Trail(BaseModel):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        trail_name: str,
        bucket_name: str,
        s3_key_prefix: str,
        sns_topic_name: str,
        is_global: bool,
        is_multi_region: bool,
        log_validation: bool,
        is_org_trail: bool,
        cw_log_group_arn: str,
        cw_role_arn: str,
        kms_key_id: str,
    ):
        self.account_id = account_id
        self.region_name = region_name
        self.trail_name = trail_name
        self.bucket_name = bucket_name
        self.s3_key_prefix = s3_key_prefix
        self.sns_topic_name = sns_topic_name
        self.is_multi_region = is_multi_region
        self.log_validation = log_validation
        self.is_org_trail = is_org_trail
        self.include_global_service_events = is_global
        self.cw_log_group_arn = cw_log_group_arn
        self.cw_role_arn = cw_role_arn
        self.kms_key_id = kms_key_id
        self.check_name()
        self.check_bucket_exists()
        self.check_topic_exists()
        self.status = TrailStatus()
        self.event_selectors: List[Dict[str, Any]] = list()
        self.advanced_event_selectors: List[Dict[str, Any]] = list()
        self.insight_selectors: List[Dict[str, str]] = list()

    @property
    def arn(self) -> str:
        return f"arn:aws:cloudtrail:{self.region_name}:{self.account_id}:trail/{self.trail_name}"

    @property
    def topic_arn(self) -> Optional[str]:
        if self.sns_topic_name:
            return f"arn:aws:sns:{self.region_name}:{self.account_id}:{self.sns_topic_name}"
        return None

    def check_name(self) -> None:
        if len(self.trail_name) < 3:
            raise TrailNameTooShort(actual_length=len(self.trail_name))
        if len(self.trail_name) > 128:
            raise TrailNameTooLong(actual_length=len(self.trail_name))
        if not re.match("^[0-9a-zA-Z]{1}.+$", self.trail_name):
            raise TrailNameNotStartingCorrectly()
        if not re.match(r".+[0-9a-zA-Z]{1}$", self.trail_name):
            raise TrailNameNotEndingCorrectly()
        if not re.match(r"^[.\-_0-9a-zA-Z]+$", self.trail_name):
            raise TrailNameInvalidChars()

    def check_bucket_exists(self) -> None:
        from moto.s3.models import s3_backends

        try:
            s3_backends[self.account_id]["global"].get_bucket(self.bucket_name)
        except Exception:
            raise S3BucketDoesNotExistException(
                f"S3 bucket {self.bucket_name} does not exist!"
            )

    def check_topic_exists(self) -> None:
        if self.topic_arn:
            from moto.sns import sns_backends

            sns_backend = sns_backends[self.account_id][self.region_name]
            try:
                sns_backend.get_topic(self.topic_arn)
            except Exception:
                raise InsufficientSnsTopicPolicyException(
                    "SNS Topic does not exist or the topic policy is incorrect!"
                )

    def start_logging(self) -> None:
        self.status.start_logging()

    def stop_logging(self) -> None:
        self.status.stop_logging()

    def put_event_selectors(
        self,
        event_selectors: List[Dict[str, Any]],
        advanced_event_selectors: List[Dict[str, Any]],
    ) -> None:
        if event_selectors:
            self.event_selectors = event_selectors
        elif advanced_event_selectors:
            self.event_selectors = []
            self.advanced_event_selectors = advanced_event_selectors

    def get_event_selectors(self) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
        return self.event_selectors, self.advanced_event_selectors

    def put_insight_selectors(self, insight_selectors: List[Dict[str, str]]) -> None:
        self.insight_selectors.extend(insight_selectors)

    def get_insight_selectors(self) -> List[Dict[str, str]]:
        return self.insight_selectors

    def update(
        self,
        s3_bucket_name: Optional[str],
        s3_key_prefix: Optional[str],
        sns_topic_name: Optional[str],
        include_global_service_events: Optional[bool],
        is_multi_region_trail: Optional[bool],
        enable_log_file_validation: Optional[bool],
        is_organization_trail: Optional[bool],
        cw_log_group_arn: Optional[str],
        cw_role_arn: Optional[str],
        kms_key_id: Optional[str],
    ) -> None:
        if s3_bucket_name is not None:
            self.bucket_name = s3_bucket_name
        if s3_key_prefix is not None:
            self.s3_key_prefix = s3_key_prefix
        if sns_topic_name is not None:
            self.sns_topic_name = sns_topic_name
        if include_global_service_events is not None:
            self.include_global_service_events = include_global_service_events
        if is_multi_region_trail is not None:
            self.is_multi_region = is_multi_region_trail
        if enable_log_file_validation is not None:
            self.log_validation = enable_log_file_validation
        if is_organization_trail is not None:
            self.is_org_trail = is_organization_trail
        if cw_log_group_arn is not None:
            self.cw_log_group_arn = cw_log_group_arn
        if cw_role_arn is not None:
            self.cw_role_arn = cw_role_arn
        if kms_key_id is not None:
            self.kms_key_id = kms_key_id

    def short(self) -> Dict[str, str]:
        return {
            "Name": self.trail_name,
            "TrailARN": self.arn,
            "HomeRegion": self.region_name,
        }

    def description(self, include_region: bool = False) -> Dict[str, Any]:
        desc = {
            "Name": self.trail_name,
            "S3BucketName": self.bucket_name,
            "IncludeGlobalServiceEvents": self.include_global_service_events,
            "IsMultiRegionTrail": self.is_multi_region,
            "TrailARN": self.arn,
            "LogFileValidationEnabled": self.log_validation,
            "IsOrganizationTrail": self.is_org_trail,
            "HasCustomEventSelectors": False,
            "HasInsightSelectors": False,
            "CloudWatchLogsLogGroupArn": self.cw_log_group_arn,
            "CloudWatchLogsRoleArn": self.cw_role_arn,
            "KmsKeyId": self.kms_key_id,
        }
        if self.s3_key_prefix is not None:
            desc["S3KeyPrefix"] = self.s3_key_prefix
        if self.sns_topic_name is not None:
            desc["SnsTopicName"] = self.sns_topic_name
            desc["SnsTopicARN"] = self.topic_arn
        if include_region:
            desc["HomeRegion"] = self.region_name
        return desc


class CloudTrailBackend(BaseBackend):
    """Implementation of CloudTrail APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.trails: Dict[str, Trail] = dict()
        self.tagging_service = TaggingService(tag_name="TagsList")

    def create_trail(
        self,
        name: str,
        bucket_name: str,
        s3_key_prefix: str,
        sns_topic_name: str,
        is_global: bool,
        is_multi_region: bool,
        log_validation: bool,
        is_org_trail: bool,
        cw_log_group_arn: str,
        cw_role_arn: str,
        kms_key_id: str,
        tags_list: List[Dict[str, str]],
    ) -> Trail:
        trail = Trail(
            self.account_id,
            self.region_name,
            name,
            bucket_name,
            s3_key_prefix,
            sns_topic_name,
            is_global,
            is_multi_region,
            log_validation,
            is_org_trail,
            cw_log_group_arn,
            cw_role_arn,
            kms_key_id,
        )
        self.trails[name] = trail
        self.tagging_service.tag_resource(trail.arn, tags_list)
        return trail

    def get_trail(self, name_or_arn: str) -> Trail:
        if len(name_or_arn) < 3:
            raise TrailNameTooShort(actual_length=len(name_or_arn))
        if name_or_arn in self.trails:
            return self.trails[name_or_arn]
        for trail in self.trails.values():
            if trail.arn == name_or_arn:
                return trail
        raise TrailNotFoundException(account_id=self.account_id, name=name_or_arn)

    def get_trail_status(self, name: str) -> TrailStatus:
        if len(name) < 3:
            raise TrailNameTooShort(actual_length=len(name))

        all_trails = self.describe_trails(include_shadow_trails=True)
        trail = next(
            (
                trail
                for trail in all_trails
                if trail.trail_name == name or trail.arn == name
            ),
            None,
        )
        if not trail:
            # This particular method returns the ARN as part of the error message
            arn = (
                f"arn:aws:cloudtrail:{self.region_name}:{self.account_id}:trail/{name}"
            )
            raise TrailNotFoundException(account_id=self.account_id, name=arn)
        return trail.status

    def describe_trails(self, include_shadow_trails: bool) -> Iterable[Trail]:
        all_trails = []
        if include_shadow_trails:
            current_account = cloudtrail_backends[self.account_id]
            for backend in current_account.values():
                for trail in backend.trails.values():
                    if trail.is_multi_region or trail.region_name == self.region_name:
                        all_trails.append(trail)
        else:
            all_trails.extend(self.trails.values())
        return all_trails

    def list_trails(self) -> Iterable[Trail]:
        return self.describe_trails(include_shadow_trails=True)

    def start_logging(self, name: str) -> None:
        trail = self.trails[name]
        trail.start_logging()

    def stop_logging(self, name: str) -> None:
        trail = self.trails[name]
        trail.stop_logging()

    def delete_trail(self, name: str) -> None:
        if name in self.trails:
            del self.trails[name]

    def update_trail(
        self,
        name: str,
        s3_bucket_name: str,
        s3_key_prefix: str,
        sns_topic_name: str,
        include_global_service_events: bool,
        is_multi_region_trail: bool,
        enable_log_file_validation: bool,
        is_organization_trail: bool,
        cw_log_group_arn: str,
        cw_role_arn: str,
        kms_key_id: str,
    ) -> Trail:
        trail = self.get_trail(name_or_arn=name)
        trail.update(
            s3_bucket_name=s3_bucket_name,
            s3_key_prefix=s3_key_prefix,
            sns_topic_name=sns_topic_name,
            include_global_service_events=include_global_service_events,
            is_multi_region_trail=is_multi_region_trail,
            enable_log_file_validation=enable_log_file_validation,
            is_organization_trail=is_organization_trail,
            cw_log_group_arn=cw_log_group_arn,
            cw_role_arn=cw_role_arn,
            kms_key_id=kms_key_id,
        )
        return trail

    def put_event_selectors(
        self,
        trail_name: str,
        event_selectors: List[Dict[str, Any]],
        advanced_event_selectors: List[Dict[str, Any]],
    ) -> Tuple[str, List[Dict[str, Any]], List[Dict[str, Any]]]:
        trail = self.get_trail(trail_name)
        trail.put_event_selectors(event_selectors, advanced_event_selectors)
        trail_arn = trail.arn
        return trail_arn, event_selectors, advanced_event_selectors

    def get_event_selectors(
        self, trail_name: str
    ) -> Tuple[str, List[Dict[str, Any]], List[Dict[str, Any]]]:
        trail = self.get_trail(trail_name)
        event_selectors, advanced_event_selectors = trail.get_event_selectors()
        return trail.arn, event_selectors, advanced_event_selectors

    def add_tags(self, resource_id: str, tags_list: List[Dict[str, str]]) -> None:
        self.tagging_service.tag_resource(resource_id, tags_list)

    def remove_tags(self, resource_id: str, tags_list: List[Dict[str, str]]) -> None:
        self.tagging_service.untag_resource_using_tags(resource_id, tags_list)

    def list_tags(self, resource_id_list: List[str]) -> List[Dict[str, Any]]:
        """
        Pagination is not yet implemented
        """
        resp: List[Dict[str, Any]] = [{"ResourceId": r_id} for r_id in resource_id_list]
        for item in resp:
            item["TagsList"] = self.tagging_service.list_tags_for_resource(
                item["ResourceId"]
            )["TagsList"]
        return resp

    def put_insight_selectors(
        self, trail_name: str, insight_selectors: List[Dict[str, str]]
    ) -> Tuple[str, List[Dict[str, str]]]:
        trail = self.get_trail(trail_name)
        trail.put_insight_selectors(insight_selectors)
        return trail.arn, insight_selectors

    def get_insight_selectors(
        self, trail_name: str
    ) -> Tuple[str, List[Dict[str, str]]]:
        trail = self.get_trail(trail_name)
        return trail.arn, trail.get_insight_selectors()


cloudtrail_backends = BackendDict(CloudTrailBackend, "cloudtrail")



import functools
from typing import TYPE_CHECKING, Any, Tuple

import requests.adapters

from moto import settings

if TYPE_CHECKING:
    from docker import DockerClient


_orig_adapter_send = requests.adapters.HTTPAdapter.send


class DockerModel:
    def __init__(self) -> None:
        self.__docker_client = None

    @property
    def docker_client(self) -> "DockerClient":  # type: ignore
        if self.__docker_client is None:
            # We should only initiate the Docker Client at runtime.
            # The docker.from_env() call will fall if Docker is not running
            import docker

            self.__docker_client = docker.from_env()

            # Unfortunately mocking replaces this method w/o fallback enabled, so we
            # need to replace it if we detect it's been mocked
            if requests.adapters.HTTPAdapter.send != _orig_adapter_send:
                _orig_get_adapter = self.docker_client.api.get_adapter

                def replace_adapter_send(*args: Any, **kwargs: Any) -> Any:
                    adapter = _orig_get_adapter(*args, **kwargs)

                    if isinstance(adapter, requests.adapters.HTTPAdapter):
                        adapter.send = functools.partial(_orig_adapter_send, adapter)  # type: ignore
                    return adapter

                self.docker_client.api.get_adapter = replace_adapter_send
        return self.__docker_client

    def ensure_image_exists(self, name: str) -> None:
        full_name = ":".join(parse_image_ref(name))
        try:
            self.docker_client.images.get(full_name)
        except:  # noqa: E722 Do not use bare except
            self.docker_client.images.pull(full_name)


def parse_image_ref(image_name: str) -> Tuple[str, str]:
    # podman does not support short container image name out of box - try to make a full name
    # See ParseDockerRef() in https://github.com/distribution/distribution/blob/main/reference/normalize.go
    parts = image_name.split("/")
    if len(parts) == 1 or (
        "." not in parts[0] and ":" not in parts[0] and parts[0] != "localhost"
    ):
        domain = settings.DEFAULT_CONTAINER_REGISTRY
        remainder = parts
    else:
        domain = parts[0]
        remainder = parts[1:]
    # Special handling for docker.io
    # https://github.com/containers/image/blob/master/docs/containers-registries.conf.5.md#normalization-of-dockerio-references
    if domain == "docker.io" and len(remainder) == 1:
        remainder = ["library"] + remainder
    if ":" in remainder[-1]:
        remainder[-1], image_tag = remainder[-1].split(":", 1)
    else:
        image_tag = "latest"
    image_repository = "/".join([domain] + remainder)
    return image_repository, image_tag



from moto.core.exceptions import JsonRESTError


class GreengrassClientError(JsonRESTError):
    code = 400


class IdNotFoundException(GreengrassClientError):
    def __init__(self, msg: str):
        self.code = 404
        super().__init__("IdNotFoundException", msg)


class InvalidContainerDefinitionException(GreengrassClientError):
    def __init__(self, msg: str):
        self.code = 400
        super().__init__("InvalidContainerDefinitionException", msg)


class VersionNotFoundException(GreengrassClientError):
    def __init__(self, msg: str):
        self.code = 404
        super().__init__("VersionNotFoundException", msg)


class InvalidInputException(GreengrassClientError):
    def __init__(self, msg: str):
        self.code = 400
        super().__init__("InvalidInputException", msg)


class MissingCoreException(GreengrassClientError):
    def __init__(self, msg: str):
        self.code = 400
        super().__init__("MissingCoreException", msg)


class ResourceNotFoundException(GreengrassClientError):
    def __init__(self, msg: str):
        self.code = 404
        super().__init__("ResourceNotFoundException", msg)



from __future__ import annotations

from typing import Final

from moto.stepfunctions.parser.asl.component.common.catch.catcher_outcome import (
    CatcherOutcomeCaught,
    CatcherOutcomeNotCaught,
)
from moto.stepfunctions.parser.asl.component.common.catch.catcher_props import (
    CatcherProps,
)
from moto.stepfunctions.parser.asl.component.common.error_name.error_equals_decl import (
    ErrorEqualsDecl,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
)
from moto.stepfunctions.parser.asl.component.common.flow.next import Next
from moto.stepfunctions.parser.asl.component.common.path.result_path import ResultPath
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class CatcherDecl(EvalComponent):
    _DEFAULT_RESULT_PATH: Final[ResultPath] = ResultPath(result_path_src="$")

    def __init__(
        self,
        error_equals: ErrorEqualsDecl,
        next_decl: Next,
        result_path: ResultPath = _DEFAULT_RESULT_PATH,
    ):
        self.error_equals: Final[ErrorEqualsDecl] = error_equals
        self.result_path: Final[ResultPath] = (
            result_path or CatcherDecl._DEFAULT_RESULT_PATH
        )
        self.next_decl: Final[Next] = next_decl

    @classmethod
    def from_catcher_props(cls, props: CatcherProps) -> CatcherDecl:
        return cls(
            error_equals=props.get(
                typ=ErrorEqualsDecl,
                raise_on_missing=ValueError(
                    f"Missing ErrorEquals declaration for Catcher declaration, in props '{props}'."
                ),
            ),
            next_decl=props.get(
                typ=Next,
                raise_on_missing=ValueError(
                    f"Missing Next declaration for Catcher declaration, in props '{props}'."
                ),
            ),
            result_path=props.get(typ=ResultPath),
        )

    @staticmethod
    def _extract_error_cause(failure_event: FailureEvent) -> dict:
        # TODO: consider formalising all EventDetails to ensure FailureEvent can always reach the state below.
        #  As per AWS's Api specification, all failure event carry one
        #  details field, with at least fields 'cause and 'error'
        specs_event_details = list(failure_event.event_details.values())
        if (
            len(specs_event_details) != 1
            and "error" in specs_event_details
            and "cause" in specs_event_details
        ):
            raise RuntimeError(
                f"Internal Error: invalid event details declaration in FailureEvent: '{failure_event}'."
            )
        spec_event_details: dict = list(failure_event.event_details.values())[0]
        error = spec_event_details["error"]
        cause = spec_event_details.get("cause") or ""
        # Stepfunctions renames these fields to capital in this scenario.
        return {
            "Error": error,
            "Cause": cause,
        }

    def _eval_body(self, env: Environment) -> None:
        failure_event: FailureEvent = env.stack.pop()

        env.stack.append(failure_event.error_name)
        self.error_equals.eval(env)

        equals: bool = env.stack.pop()
        if equals:
            error_cause: dict = self._extract_error_cause(failure_event)
            env.stack.append(error_cause)

            self.result_path.eval(env)
            env.next_state_name = self.next_decl.name
            env.stack.append(CatcherOutcomeCaught())
        else:
            env.stack.append(failure_event)
            env.stack.append(CatcherOutcomeNotCaught())
